from typing import Dict, Any, List, Optional, Callable, Union

import armarx.armem as armem

from armarx_memory.core import MemoryID, time_usec
from armarx_memory.client.Commit import Commit, EntityUpdate


class Writer:

    WritingMemoryServerPrx = "armarx.armem.server.WritingMemoryInterfacePrx"

    def __init__(
        self,
        server: Optional[WritingMemoryServerPrx] = None,
    ):
        self.server = server

    def add_provider_segment(
        self,
        provider_id: MemoryID,
        clear_when_exists=False,
    ):

        inp = armem.data.AddSegmentInput()
        inp.coreSegmentName = provider_id.core_segment_name
        inp.providerSegmentName = provider_id.provider_segment_name
        inp.clearWhenExists = clear_when_exists
        results = self.server.addSegments([inp])
        return results[0]

    def commit(
        self,
        commit: Commit,
    ):

        time_sent = time_usec()
        for update in commit.updates:
            update.time_sent_usec = time_sent

        ice_commit = commit.to_ice()

        ice_result = self.server.commit(ice_commit)

        return ice_result

    def __bool__(self):
        return bool(self.server)



from argparse import ArgumentParser

import sys
import logging

from .ice_manager import test_connection


class ArmarXArgumentParser(ArgumentParser):
    """
    ArgumentParser with default arguments

    ..see:: argparse.ArgumentParser

    ..highlight:: python
    ..codeblock:: python

        from armarx.parser import ArmarXArgumentParser as ArgumentParser

        parser = ArgumentParser('test')
        parser.add_argument(...)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def parse_args(self):
        self.add_argument(
            "--skip-connection-test",
            action="store_true",
            help="Do not test the connection",
        )
        verbose_group = self.add_mutually_exclusive_group()
        verbose_group.add_argument(
            "-v", "--verbose", action="store_true", help="be verbose"
        )
        verbose_group.add_argument(
            "-q", "--quiet", action="store_true", help="be quiet"
        )

        args = super().parse_args()

        log_format = "%(asctime)s - %(levelname)s - %(message)s"
        if args.verbose:
            logging.basicConfig(
                format=log_format, stream=sys.stdout, level=logging.DEBUG
            )
        elif args.quiet:
            logging.basicConfig(
                format=log_format, stream=sys.stdout, level=logging.ERROR
            )
        else:
            logging.basicConfig(
                format=log_format, stream=sys.stdout, level=logging.INFO
            )

        if not args.skip_connection_test:
            test_connection()

        return args



import numpy as np


def kit_colors():
    """

    https://intranet.kit.edu/gestaltungsrichtlinien.php
    """
    color_names = [
        "Grün",
        "Blau",
        "Gelb",
        "Orange",
        "Maigrün",
        "Rot",
        "Lila",
        "Braun",
        "Cyan",
    ]
    kit_colors = np.array(
        [
            [0, 150, 130, 255],
            [70, 100, 170, 255],
            [252, 229, 0, 255],
            [223, 155, 27, 255],
            [140, 182, 60, 255],
            [162, 34, 35, 255],
            [163, 16, 124, 255],
            [167, 130, 46, 255],
            [35, 161, 224, 255],
        ]
    )
    kit_colors = kit_colors / 255
    return zip(color_names, kit_colors)


def kit_color_map():
    from matplotlib.colors import ListedColormap

    return ListedColormap(kit_colors)



import os
import numpy as np
import mmap
from armarx_core.config import config


def path_to_shm(provider_name: str) -> str:
    if not config.getboolean("Misc", "python_shm_support", fallback=False):
        return None
    username = os.getlogin()
    shm_file = f"/dev/shm/{provider_name}MemoryImageProvider{username}"
    if os.path.isfile(shm_file):
        return shm_file
    return None


def read_images_shm(shm_file: str, data_dimensions):
    image_size = np.prod(data_dimensions)
    shm_size = image_size + 1024 + 128
    with open(shm_file) as f:
        with mmap.mmap(f.fileno(), length=shm_size, access=mmap.ACCESS_READ) as m:
            data = m.read()

    image = np.frombuffer(data, dtype=np.uint8)
    image = image[696:]
    image = image[:image_size]
    image = image.reshape(data_dimensions)

    return image



import socket
import time

from armarx.ice_conv.ice_converter import IceConverter
from armarx_core import slice_loader

slice_loader.load_armarx_slice("ArmarXCore", "core/time.ice")
from armarx.core.time.dto import DateTime
from armarx.core.time.dto.ClockType import ClockTypeEnum


def time_usec() -> int:
    return int(time.time() * 1e6)


class DateTimeIceConverter(IceConverter):
    @classmethod
    def _import_dto(cls):
        return DateTime

    def _from_ice(self, dto: DateTime) -> int:
        return dto.timeSinceEpoch.microSeconds

    def _to_ice(self, bo: int) -> DateTime:
        dto = DateTime()
        dto.timeSinceEpoch.microSeconds = bo
        dto.clockType = ClockTypeEnum.Monotonic
        dto.hostname = socket.gethostname()
        return dto


if __name__ == "__main__":
    c = DateTimeIceConverter()
    bo = time_usec()
    dto = c.to_ice(bo)
    bo2 = c.from_ice(dto)



import numpy as np

from typing import Union, List


def spherical2cartesian(
    spherical: Union[List, np.ndarray],
) -> np.ndarray:

    spherical = np.array(spherical)
    assert spherical.shape[-1] == 3
    radius = spherical[..., 0]
    azim = spherical[..., 1]
    elev = spherical[..., 2]
    inclination = np.pi / 2 - elev
    sin_inclination = np.sin(inclination)

    cartesian = spherical.astype(np.float).copy()
    cartesian[..., 0] = radius * sin_inclination * np.cos(azim)  # x
    cartesian[..., 1] = radius * sin_inclination * np.sin(azim)  # y
    cartesian[..., 2] = radius * np.cos(inclination)  # z
    return cartesian


def cartesian2spherical(
    cartesian: Union[List, np.ndarray],
) -> np.ndarray:

    cartesian = np.array(cartesian)
    assert cartesian.shape[-1] == 3
    x = cartesian[..., 0]
    y = cartesian[..., 1]
    z = cartesian[..., 2]

    radius = np.linalg.norm(cartesian, axis=-1)  # radius
    azim = np.arctan2(y, x)  # angle phi
    inclination = np.arccos(z / radius)  # angle theta
    elev = np.pi / 2 - inclination

    polar = cartesian.astype(np.float).copy()
    polar[..., 0] = radius
    polar[..., 1] = azim
    polar[..., 2] = elev
    return polar



from typing import List, Union, Tuple


from armarx.viz.data import Element
from armarx.viz.data import LayerUpdate

from .elements import Element as E


class Layer:
    def __init__(self, component: str, name: str, elements: List[E] = ()):
        self.component = component
        self.name = name
        self.elements: List[E] = list(elements)

    def clear(self):
        self.elements = []

    def add(self, element: Union[Element, E]):
        self.elements.append(element)

    @property
    def id(self) -> str:
        return "{}/{}".format(self.component, self.name)

    def data(self) -> LayerUpdate:
        return LayerUpdate(
            component=self.component,
            name=self.name,
            elements=list(map(self._get_element_data, self.elements)),
        )

    @staticmethod
    def _get_element_data(element: Union[Element, E]):
        if isinstance(element, E):
            return element.get_ice_data()
        elif isinstance(element, Element):
            return element
        else:
            raise TypeError(
                "Expected viz element, but received object of class '{}'.".format(
                    element.__class__
                )
            )

    def __iadd__(self, elements):
        """Add an element or multiple elements to this layer's elements."""
        if isinstance(elements, E) or isinstance(elements, Element):
            self.elements.append(elements)
        else:
            self.elements += elements
        return self

    def __enter__(self):
        """
        Makes a layer usable in a with statement:
        with arviz.stage("MyLayer") as layer:
            layer.elements += [ ... ]
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import dataclasses as dc
import typing as ty

from armarx_memory.aron.aron_dataclass import AronDataclass


@dc.dataclass
class PackagePath(AronDataclass):

    package: str
    path: str

    def get_system_path(self) -> str:
        import os
        from armarx import cmake_helper

        [data_path] = cmake_helper.get_data_path(self.package)
        abs_path = os.path.join(data_path, self.package, self.path)
        return abs_path



import logging

import os
import subprocess
from functools import lru_cache

from typing import List

logger = logging.getLogger(__name__)


@lru_cache(maxsize=32)
def get_armarx_include_dirs(pkg_name: str) -> List[str]:
    """
    find the include dir path for an armarx package

    :param pkg_name: name of the package
    :returns: the include path to the package if found
    :rtype: str
    """
    cmd = [
        "cmake",
        "--find-package",
        f"-DNAME={pkg_name}",
        "-DCOMPILER_ID=GNU",
        "-DLANGUAGE=C",
        "-DMODE=COMPILE",
    ]
    result = subprocess.check_output(cmd).decode("utf-8")
    includes = []
    path_list = result.split("-I")
    for path in path_list:
        if path.strip():
            includes.append(path.strip())
    return includes


@lru_cache(maxsize=32)
def get_data_path(package_name: str) -> List[str]:
    return get_package_information(package_name, "DATA_DIR:")


def get_dependencies(package_name: str, include_self=False) -> List[str]:
    dependencies = get_package_information(package_name, "SOURCE_PACKAGE_DEPENDENCIES:")
    dependencies = dependencies or []
    # manually add ArmarXCore
    if not "ArmarXCore" in dependencies:
        dependencies.append("ArmarXCore")
    if include_self and is_armarx_package(package_name):
        dependencies.append(package_name)
        return dependencies
    else:
        return dependencies or []


@lru_cache(maxsize=32)
def get_include_path(package_name: str) -> List[str]:
    return get_package_information(package_name, "INTERFACE_DIRS:")


def get_build_path(package_name: str) -> List[str]:
    return get_package_information(package_name, "BUILD_DIR:")


def get_package_information(package_name: str, info) -> List[str]:
    package_data = get_package_data(package_name)
    for l in package_data.split("\n"):
        if info in l:
            if l.endswith(":"):
                return []
            l = l.split(":")[1]
            return l.split(";")


@lru_cache(maxsize=32)
def get_package_data(package_name: str):
    if not package_name:
        logger.error("package name is empty.")
        return
    rel_cmake_script = "ArmarXCore/core/system/cmake/FindPackageX.cmake"
    includes = get_armarx_include_dirs("ArmarXCore")
    for include in includes:
        cmake_script = os.path.join(include, rel_cmake_script)
        if os.path.exists(cmake_script):
            armarx_cmake_script = cmake_script
            cmd = [
                "cmake",
                "-DPACKAGE={}".format(package_name),
                "-P",
                armarx_cmake_script,
            ]
            return subprocess.check_output(cmd).decode("utf-8")
    logger.error("Could not find %s for ArmarXCore in %s", rel_cmake_script, includes)
    raise ValueError("Could not find a valid ArmarXCore path!")


def is_armarx_package(package_name: str) -> bool:
    package_data = get_package_data(package_name)
    return package_data.strip()


@lru_cache(maxsize=32)
def find_cmake_package(
    package_name: str,
    package_cache_dir: str = "~/.cmake/packages",
) -> str:
    """
    Find a CMake package using the package cache.

    :param package_name: The name of the CMake project.
    :param package_cache_dir: The directory of the cmake package cache.
    :return: The absolute path to the package root.
    :raise: IOError If the package cannot be found.
    """
    cmake_package_cache_dir = os.path.expanduser(package_cache_dir)
    directory = os.path.join(cmake_package_cache_dir, package_name)

    if not os.path.isdir(directory):
        msg = "Could not find CMake package '{}' in CMake package cache '{}'.".format(
            package_name, cmake_package_cache_dir
        )

        # Case mismatch?
        dirs = sorted(os.listdir(cmake_package_cache_dir))
        dirs_lower = [d.lower() for d in dirs]
        try:
            index = dirs_lower.index(package_name.lower())
            msg += f"\nDid you mean '{dirs[index]}'?"
        except ValueError:
            pass

        raise IOError(msg)

    files = os.listdir(directory)
    if len(files) == 1:
        with open(os.path.join(directory, files[0])) as file:
            content = file.read()
        build_dir = content.strip()
        root_dir, build = os.path.split(build_dir)
        if build != "build":
            raise IOError("Something strange")
        return root_dir

    elif len(files) == 0:
        msg = f"Found no file in directory: {directory}"
        raise IOError(msg)

    else:
        msg = "Found more than one file with the following contents:"
        for filename in sorted(files):
            with open(os.path.join(directory, filename)) as file:
                content = file.read().strip()
            msg += f"\n- {file}: \t{content}"
        raise IOError(msg)



"""
Convenience module to convert armarx.FramedPose and armarx.FramedPosition to
numpy arrays and vice versa.

Please note that the numpy array's data type must be np.float32
"""

import numpy as np
import transforms3d as tf3d

from armarx import slice_loader

slice_loader.load_armarx_slice("ArmarXCore", "components/EmergencyStopInterface.ice")
slice_loader.load_armarx_slice(
    "ArmarXCore", "components/SimpleStatechartExecutorInterface.ice"
)

from armarx import RobotStateComponentInterfacePrx
from armarx import FramedPositionBase
from armarx import FramedPoseBase
from armarx import FramedOrientationBase

from armarx import PoseBase
from armarx import Vector3Base
from armarx import QuaternionBase


def mat2pose(pose: np.ndarray, frame: str = None, agent: str = None) -> FramedPoseBase:
    q = tf3d.quaternions.mat2quat(pose[:3, :3])
    q = QuaternionBase(*q)
    v = Vector3Base(*pose[:3, 3])
    if not frame:
        return PoseBase(v, q)
    else:
        return FramedPoseBase(v, q, frame=frame, agent=agent)


def pose2mat(pose: FramedPoseBase) -> np.ndarray:
    """
    Converts a FramedPoseBase to a homogeneous matrix

    Does not change the coordinate system

    :param pose: FramedPoseBase
    :return: numpy.ndarry
    """
    qw = pose.orientation.qw
    qx = pose.orientation.qx
    qy = pose.orientation.qy
    qz = pose.orientation.qz
    rot_mat = tf3d.quaternions.quat2mat([qw, qx, qy, qz])
    transform_mat = np.identity(4, dtype=np.float32)
    transform_mat[0:3, 0:3] = rot_mat
    position = pose.position
    transform_mat[0, 3] = position.x
    transform_mat[1, 3] = position.y
    transform_mat[2, 3] = position.z
    return transform_mat


def convert_position_to_global(f: FramedPositionBase) -> np.ndarray:
    pose = FramedPoseBase(
        position=f, orientation=FramedOrientationBase(), frame=f.frame, agent=f.agent
    )
    return convert_pose_to_global(pose)


def convert_mat_to_global(pose: np.ndarray, frame: str) -> np.ndarray:
    robot_state = RobotStateComponentInterfacePrx.get_proxy()
    current_robot_state = robot_state.getSynchronizedRobot()
    robot_pose = current_robot_state.getGlobalPose()
    robot_node = current_robot_state.getRobotNode(frame).getPoseInRootFrame()
    transform_robot_node_to_root = pose2mat(robot_node)
    transform_root_to_global = pose2mat(robot_pose)


def inv(pose: np.ndarray) -> np.ndarray:
    """
    can be also done by np.linalg.inv but this is a little bit faster
    """
    inv_pose = pose.copy()
    inv_pose[:3, :3] = pose[:3, :3].T
    inv_pose[:3, 3] = -1 * np.dot(pose[:3, :3].T, pose[:3, 3])
    return inv_pose


def robot_state(timestamp: int = None):
    """
    Convenience method to get the robot state. If not timestamp is given return
    the current state
    """
    robot_state = RobotStateComponentInterfacePrx.get_proxy()
    if timestamp:
        return robot_state.getRobotSnapshotAtTimestamp(timestamp)
    return robot_state.getSynchronizedRobot()


def convert_mat_to_robot_node(
    pose: np.ndarray, frame: str, timestamp: int = None
) -> np.ndarray:
    """
    given a global pose

    inverse of convert_mat_to_global
    """
    current_robot_state = robot_state(timestamp)

    robot_node_pose = pose2mat(current_robot_state.getRobotNode(frame).getGlobalPose())
    return np.dot(inv(robot_node_pose), pose)


def convert_mat_to_global(
    pose: np.ndarray, frame: str, timestamp: int = None
) -> np.ndarray:
    current_robot_state = robot_state(timestamp)
    robot_node = current_robot_state.getRobotNode(frame)
    robot_node_pose = pose2mat(robot_node.getGlobalPose())
    return np.dot(robot_node_pose, pose)


def convert_mat_to_root(
    pose: np.ndarray, frame: str, timestamp: int = None
) -> np.ndarray:
    current_robot_state = robot_state(timestamp)
    if frame == "Global" or frame == "armarx::Global":
        robot_pose = pose2mat(current_robot_state.getGlobalPose())
        return np.dot(robot_pose, pose)
    else:
        robot_node = current_robot_state.getRobotNode(frame).getPoseInRootFrame()
        transform_robot_node_to_root = pose2mat(robot_node)
        return np.dot(transform_robot_node_to_root, pose)


def convert_pose_to_global(f: FramedPoseBase) -> np.ndarray:
    """
    Converts a armarx.FramedPoseBase to a numpy array in the global coordinate
    system

    """
    transform = pose2mat(f)
    if f.frame == "Global" or f.frame == "armarx::Global":
        return transform
    return convert_mat_to_global(transform, f.frame)


def convert_pose_to_root(f: FramedPoseBase) -> np.ndarray:
    """
    Converts a armarx.FramedPoseBase to a numpy array
    """
    transform = pose2mat(f)
    return convert_mat_to_root(transform, f.frame)



from typing import Optional

from armarx_core import slice_loader

slice_loader.load_armarx_slice("RobotAPI", "aron.ice")
slice_loader.load_armarx_slice("RobotAPI", "armem/memory.ice")

from armarx import aron
from armarx import armem


def get_latest_data(
    entity: armem.data.Entity,
    instance_index=0,
) -> "aron.AronData":
    snapshot: armem.data.EntitySnapshot = list(entity.history.values())[-1]
    instance: armem.data.EntityInstance = snapshot.instances[instance_index]
    return instance.data


def find_entity_by_name(
    name: str,
    core_segment: armem.data.CoreSegment,
) -> Optional[armem.data.Entity]:
    # Find class
    for prov_name, prov in core_segment.providerSegments.items():
        for entity_name, entity in prov.entities.items():
            if entity_name == name:
                return entity
    return None



from armarx_robots.armar_6 import A6
from armarx_robots.armar_de import AD
from armarx_robots.armar_3 import A3


def robot_by_name(robot_name: str = None):
    if robot_name == "A6":
        return A6()
    elif robot_name == "AD":
        return AD()
    elif robot_name == "A3":
        return A3()
    return None



import numpy as np

from .import_aron_slice import import_aron_slice


dtype_rgb = [("r", "i1"), ("g", "i1"), ("b", "i1")]

dtypes_dict = {
    "float": np.float32,
    "float32": np.float32,
    "double": np.float64,
    "float64": np.float64,
    "16": dtype_rgb,  # "16" == OpenCV 8UC3 = RGB image
    # "16": np.float32,  # "16" == OpenCV F1C1 = Depth image
}


def convert_dtype_rgb_to_int8(array: np.ndarray) -> np.ndarray:
    """
    Converts an array with shape (m, n) and dtype dtype_rgb
    to an array with shape (m, n, 3) and dtype int8.

    :param array: The RGB image with structured dtype.
    :return: The RGB image with native dtype.
    """
    return np.stack([array[c] for c in "rgb"], axis=-1)


try:
    # beta 0.2.3
    class AronIceTypes:
        ARON_VERSION = "beta 0.2.3"

        ns = import_aron_slice().data.dto

        Data = ns.GenericData

        String = ns.AronString
        Bool = ns.AronBool
        Int = ns.AronInt
        Long = ns.AronLong
        Float = ns.AronFloat

        List = ns.List
        Dict = ns.Dict

        NDArray = ns.NDArray

        @classmethod
        def string(cls, value: str) -> String:
            return cls.String(value=value, VERSION=cls.ARON_VERSION)

        @classmethod
        def bool(cls, value: int) -> Bool:
            return cls.Bool(value=value, VERSION=cls.ARON_VERSION)

        @classmethod
        def int(cls, value: int) -> Int:
            return cls.Int(value=value, VERSION=cls.ARON_VERSION)

        @classmethod
        def long(cls, value: int) -> Long:
            return cls.Long(value=value, VERSION=cls.ARON_VERSION)

        @classmethod
        def float(cls, value: float) -> Float:
            return cls.Float(value=value, VERSION=cls.ARON_VERSION)

        @classmethod
        def list(cls, elements: list) -> List:
            return cls.List(elements=elements, VERSION=cls.ARON_VERSION)

        @classmethod
        def dict(cls, elements: dict) -> Dict:
            return cls.Dict(elements=elements, VERSION=cls.ARON_VERSION)

except AttributeError as e:
    # < 0.2.3

    class AronIceTypes:

        ns = import_aron_slice().data

        Data = ns.AronData

        String = ns.AronString
        Bool = ns.AronBool
        Int = ns.AronInt
        Long = ns.AronLong
        Float = ns.AronFloat

        List = ns.AronList
        Dict = ns.AronDict

        NdArray = ns.AronNDArray



import enum
from typing import Dict, Tuple

import numpy as np


import armarx.viz as viz
from armarx.arviz import conversions
from armarx.arviz.elements.Element import Element


def direction_to_ori_mat(dir: np.ndarray, natural_dir=(0, 1, 0)) -> np.ndarray:
    import transforms3d as tf3d

    dir = np.array(dir)
    dir /= np.linalg.norm(dir)
    cross = np.cross(natural_dir, dir)
    angle = np.arccos(natural_dir.dot(dir))
    if np.linalg.norm(angle) < 1e-6:
        # Directions are almost colinear => Do no rotation
        cross = np.array((1, 0, 0))
        angle = 0.0
    axis = cross / np.linalg.norm(cross)
    ori = tf3d.axangles.axangle2mat(axis, angle)
    return ori


class Arrow(Element):
    """
    An arrow.
    """

    natural_dir = np.array((0, 1, 0))

    def __init__(
        self,
        id,
        length=100.0,
        width=10.0,
        direction=None,
        from_to=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementArrow, id=id, **kwargs)
        self.length: float = length
        self.width: float = width
        if direction is not None:
            self.direction = direction
        if from_to is not None:
            self.from_to = from_to

    @property
    def direction(self):
        return self.ori_mat @ self.natural_dir

    @direction.setter
    def direction(self, value):
        self.ori_mat = direction_to_ori_mat(value, self.natural_dir)

    @property
    def from_to(self):
        return self.position, self.position + self.length * self.direction

    @from_to.setter
    def from_to(self, value):
        start, end = value
        self.position = start
        dir = end - self.position
        norm = np.linalg.norm(dir)
        self.direction = dir / norm
        self.length = norm

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.length = float(self.length)
        ice_data.width = float(self.width)


class ArrowCircle(Element):
    """
    An arrow circle.
    """

    natural_normal = np.array((0, 1, 0))

    def __init__(
        self,
        id,
        radius=100.0,
        completion=1.0,
        width=10.0,
        normal=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementArrowCircle, id=id, **kwargs)
        self.radius: float = radius
        self.completion: float = completion
        self.width: float = width
        if normal is not None:
            self.normal = normal

    @property
    def normal(self) -> np.ndarray:
        return self.ori_mat @ self.natural_normal

    @normal.setter
    def normal(self, value):
        self.ori_mat = direction_to_ori_mat(value, self.natural_normal)

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.radius = float(self.radius)
        ice_data.completion = float(self.completion)
        ice_data.width = float(self.width)


class Box(Element):
    """
    A box.
    """

    def __init__(
        self,
        id,
        size=1.0,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementBox, id=id, **kwargs)

        self.size = size

    @property
    def size(self) -> np.ndarray:
        return self._size

    @size.setter
    def size(self, value):
        try:
            iter(value)
            value = self._to_array_checked(value, (3,), "size vector", np.float)
            self._size = value
        except TypeError:
            self._size = np.array([value, value, value]).astype(np.float)

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.size = conversions.vector3f_from_numpy(self.size)


class Cylinder(Element):
    """
    A cylinder.
    """

    natural_dir = np.array((0, 1, 0))

    def __init__(
        self,
        id,
        radius=10.0,
        height=10.0,
        direction=None,
        from_to=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementCylinder, id=id, **kwargs)
        self.radius: float = radius
        self.height: float = height

        if direction is not None:
            self.direction = direction
        if from_to is not None:
            self.from_to = from_to

    @property
    def from_to(self):
        pos = self.position
        dir = self.direction
        return pos - self.height / 2 * dir, pos + self.height / 2 * dir

    @from_to.setter
    def from_to(self, value):
        start, end = value
        start, end = np.array(start), np.array(end)
        self.position = 0.5 * (start + end)
        direction = end - start
        self.direction = direction
        self.height = np.linalg.norm(direction)

    @property
    def direction(self):
        return self.ori_mat @ self.natural_dir

    @direction.setter
    def direction(self, value):
        self.ori_mat = direction_to_ori_mat(value, self.natural_dir)

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.radius = self.radius
        ice_data.height = self.height


class Sphere(Element):
    """
    A sphere.
    """

    def __init__(
        self,
        id,
        radius=10.0,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementSphere, id=id, **kwargs)
        self.radius: float = radius

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.radius = float(self.radius)


class Ellipsoid(Element):
    """
    An ellipsoid.
    """

    def __init__(
        self,
        id,
        axis_lengths=None,
        curvature=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementEllipsoid, id=id, **kwargs)

        self.axis_lengths = (1, 1, 1) if axis_lengths is None else axis_lengths
        self.curvature = (0, 0, 0) if curvature is None else curvature

    @property
    def axis_lengths(self) -> np.ndarray:
        return self._axis_lengths

    @axis_lengths.setter
    def axis_lengths(self, value):
        try:
            iter(value)
            value = self._to_array_checked(
                value, (3,), "ellipsoid axis lengths", np.float
            )
            self._axis_lengths = value
        except TypeError:
            self._axis_lengths = np.array([value, value, value]).astype(np.float)

    @property
    def curvature(self) -> np.ndarray:
        return self._curvature

    @curvature.setter
    def curvature(self, value):
        value = self._to_array_checked(value, (3,), "curvature", np.float)
        self._curvature = value

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.axisLengths = conversions.vector3f_from_numpy(self.axis_lengths)
        ice_data.curvature = conversions.vector3f_from_numpy(self.curvature)


class Line(Element):
    """
    A line.
    """

    def __init__(
        self,
        id,
        start=None,
        end=None,
        line_width=10.0,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementLine, id=id, **kwargs)

        self.line_width: float = line_width

        if start is not None:
            self.start = start
        if end is not None:
            self.end = end

    @property
    def start(self) -> np.ndarray:
        return self._start

    @start.setter
    def start(self, value):
        value = self._to_array_checked(value, (3,), "start position")
        self._start = value

    @property
    def end(self) -> np.ndarray:
        return self._end

    @end.setter
    def end(self, value):
        value = self._to_array_checked(value, (3,), "start position")
        self._end = value

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data._from = conversions.vector3f_from_numpy(self.start)
        ice_data.to = conversions.vector3f_from_numpy(self.end)


class Pose(Element):
    """
    A 6-D pose.
    """

    def __init__(
        self,
        id,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementPose, id=id, **kwargs)


class Text(Element):
    def __init__(
        self,
        id,
        text="",
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementText, id=id, **kwargs)
        self.text: str = text

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.text = str(self.text)


class ModelDrawStyle(enum.IntFlag):
    ORIGINAL = 0
    COLLISION = 1
    OVERRIDE_COLOR = 2


class Object(Element):
    def __init__(
        self,
        id,
        project="",
        filename="",
        file=None,
        use_collision_model=False,
        override_color=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementObject, id=id, **kwargs)
        self.project: str = project
        self.filename: str = filename
        self.draw_style: ModelDrawStyle = ModelDrawStyle.ORIGINAL

        if file is not None:
            self.file = file

        if use_collision_model:
            self.use_collision_model()
        if override_color is not None:
            self.override_color(override_color)

    @property
    def file(self) -> Tuple[str, str]:
        return self.project, self.filename

    @file.setter
    def file(self, value):
        self.project, self.filename = value

    @property
    def draw_style(self) -> ModelDrawStyle:
        return self._draw_style

    @draw_style.setter
    def draw_style(self, value):
        self._draw_style = ModelDrawStyle(value)

    def use_collision_model(self):
        self.draw_style |= ModelDrawStyle.COLLISION

    def use_full_model(self):
        self.draw_style &= ~ModelDrawStyle.COLLISION

    def override_color(self, color):
        self.draw_style |= ModelDrawStyle.OVERRIDE_COLOR
        self.color = color

    def use_original_color(self):
        self.draw_style &= ~ModelDrawStyle.OVERRIDE_COLOR

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.project = str(self.project)
        ice_data.filename = str(self.filename)
        ice_data.drawStyle = int(self.draw_style)


class Robot(Object):
    """
    A robot.
    """

    def __init__(
        self,
        id,
        joint_angles=None,
        **kwargs,
    ):
        super().__init__(id=id, **kwargs)
        self.ice_data_cls = viz.data.ElementRobot

        self.joint_angles: Dict[str, float] = (
            {} if joint_angles is None else joint_angles
        )

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.jointValues = self.joint_angles


class PointCloud(Element):
    """
    A point cloud.
    """

    def __init__(
        self,
        id,
        transparency=0.0,
        point_size=1.0,
        points=None,
        point_colors=None,
        **kwargs,
    ):
        """
        :param id:
        :param transparency:
        :param point_size: The point size in pixels.
        :param points:
        :param point_colors:
        :param kwargs:
        """
        super().__init__(ice_data_cls=viz.data.ElementPointCloud, id=id, **kwargs)
        self.points = np.zeros((0, 7))

        self.transparency: float = transparency
        self.point_size: float = point_size

        if points is not None:
            self.points = points
        if point_colors is not None:
            self.point_colors = point_colors

    def clear(self):
        self.points = []

    @property
    def points(self) -> np.ndarray:
        """
        An array of shape (N, 7) containing N points of the form (x, y, z, r, g, b, a),
        or an empty array if there are no points.
        """
        return self._points

    @points.setter
    def points(self, value):
        """
        :param value:
        An array-like of one of the following shapes:
            (N, 3): Set as (x, y, z) with default color (100, 100, 100, 255).
            (N, 6): Set as (x, y, z, r, g, b) with default alpha (255).
            (N, 7): Set as (x, y, z, r, g, b, a).
        """
        value = self._to_array_checked(
            value, [(0,), (None, 3), (None, 6), (None, 7)], "points", dtype=np.float
        )
        if value.size == 0:
            self._points = value
            return

        self._points = np.stack([[0, 0, 0, 100, 100, 100, 255]] * value.shape[0])
        self._points[:, : value.shape[-1]] = value

    @property
    def point_positions(self) -> np.ndarray:
        """An N x 3 slice of `self.points` containing the point positions as (x, y, z)."""
        return self._points[:, :3]

    @point_positions.setter
    def point_positions(self, value):
        value = self._to_array_checked(
            value, [(None, 3)], "point positions", dtype=np.float
        )
        self._points[:, :3] = value

    @property
    def point_colors(self) -> np.ndarray:
        """An N x 4 slice of `self.points` containing the point colors as (r, g, b, a)."""
        return self._points[:, 3:]

    @point_colors.setter
    def point_colors(self, value):
        value = self._to_array_checked(
            value, [(3,), (4,), (None, 3), (None, 4)], "point colors"
        )
        self._points[:, 3 : (3 + value.shape[-1])] = value

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)

        dtype = np.dtype(
            [
                ("position", np.float32, (3,)),
                ("a", np.uint8),
                ("r", np.uint8),
                ("g", np.uint8),
                ("b", np.uint8),
            ]
        )
        buffer = np.zeros(self.points.shape[0], dtype=dtype)

        assert self.points.shape[1] == 7
        buffer["position"] = self.points[:, :3]
        buffer["r"] = self.points[:, 3]
        buffer["g"] = self.points[:, 4]
        buffer["b"] = self.points[:, 5]
        buffer["a"] = self.points[:, 6]

        ice_data.points = buffer
        ice_data.transparency = self.transparency
        ice_data.point_size = self.point_size


class Polygon(Element):
    """
    A polygon.
    """

    def __init__(
        self,
        id,
        line_width=0.0,
        line_color=None,
        points=None,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementPolygon, id=id, **kwargs)
        self.points = []

        self.line_width: float = line_width
        self.line_color = line_color if line_color is not None else (100, 100, 100, 255)

        if points is not None:
            self.points = points

    @property
    def line_color(self) -> np.ndarray:
        """
        The line color of the polygon
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        value = self._to_array_checked(value, [(3,), (4,)], "line color")
        if value.shape == (3,):
            self._line_color = np.concatenate([value, [255]])
        else:
            self._line_color = value

    def clear(self):
        self.points = []

    @property
    def points(self) -> np.ndarray:
        return self._points

    @points.setter
    def points(self, value):
        value = self._to_array_checked(
            value, [(0,), (None, 3)], "polygon points", dtype=np.float
        )
        self._points = value

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)
        ice_data.line_width = float(self.line_width)
        ice_data.line_color = conversions.to_viz_color(self.color)
        ice_data.points = conversions.vector3fs_from_numpy(self.points)



from typing import Iterable, Union, List

import numpy as np


class Transform:
    def __init__(
        self,
        transform=None,
        translation=None,
        rotation=None,
    ):

        self.transform = np.eye(4)
        if transform is not None:
            self.transform = transform
        if translation is not None:
            self.translation = translation
        if rotation is not None:
            self.rotation = rotation

    @property
    def transform(self) -> np.ndarray:
        """The transform as 4x4 transformation matrix."""
        return self._transform

    @transform.setter
    def transform(self, value):
        value = self._to_array_checked(
            value, (4, 4), "transform matrix", dtype=np.float
        )
        self._transform = value

    @property
    def translation(self) -> np.ndarray:
        return self._transform[:3, 3]

    @translation.setter
    def translation(self, value):
        value = self._to_array_checked(
            value, [(3,), (3, 1)], "translation vector", dtype=np.float
        )
        self._transform[:3, 3] = value

    @property
    def rotation(self):
        return self.rot_mat

    @rotation.setter
    def rotation(self, value):
        value = np.array(value)
        if value.shape == (4,):
            self.rot_quat = value
        elif value.shape == (3, 3):
            self.rot_mat = value
        else:
            raise ValueError(f"Invalid rotation value: {value}")

    @property
    def rot_mat(self) -> np.ndarray:
        """The rotation as 3x3 rotation matrix."""
        return self._transform[:3, :3]

    @rot_mat.setter
    def rot_mat(self, value):
        """Set the rotation as 3x3 rotation matrix."""
        value = self._to_array_checked(value, (3, 3), "rotation matrix", dtype=np.float)
        self._transform[:3, :3] = value

    @property
    def rot_quat(self):
        """The rotation as [w, x, y, z] quaternion."""
        import transforms3d as tf3d

        try:
            return tf3d.quaternions.mat2quat(self.rot_mat)
        except np.linalg.LinAlgError as e:
            print("Rotation matrix to quaternion: \n{}".format(self.rot_mat))
            raise

    @rot_quat.setter
    def rot_quat(self, value):
        import transforms3d as tf3d

        try:
            self.rot_mat = tf3d.quaternions.quat2mat(value)
        except np.linalg.LinAlgError as e:
            print("Quaternion to rotation matrix: \n{}".format(value))
            raise

    def inverted(self) -> "Transform":
        """
        Return the inverted transform of self.
        :return: self^(-1)
        """
        return Transform(
            rotation=self.rot_mat.T, translation=-self.rot_mat.T @ self.translation
        )

    def __mul__(self, other: "Transform"):
        """
        Return (self * other) with * being the matrix multiplication.
        :param other: Another transform.
        :return: self * other
        """
        if not isinstance(other, Transform):
            raise TypeError(
                f"Unsupported operator * between {type(self)} and {type(other)}."
            )
        return Transform(self.transform @ other.transform)

    def is_close_to(self, other: "Transform", **kwargs):
        """
        Return true of other is close to self.
        :param other: Another transform.
        :param kwargs: Keyword arguments passed to np.allclose()
        :return: self == other
        """
        return np.allclose(self.transform, other.transform, **kwargs)

    @classmethod
    def _match_shape(
        cls, shape: Iterable[int], shape_pattern: Iterable[Union[int, None]]
    ):
        """
        Indicate whether `shape` matches `accepted_shape`.
        :param shape: An array shape, i.e. tuple of ints.
        :param shape_pattern:
            A shape pattern, i.e. tuple of ints or None, where None matches any size.
        :return: True if `shape` matches `accepted_shape`.
        """
        return np.all(
            np.logical_or(
                np.array(shape) == shape_pattern,
                np.isnan(np.array(shape_pattern, dtype=np.float)),
            )
        )

    @classmethod
    def _match_shapes(
        cls, shape: Iterable[int], shape_patterns: List[Iterable[Union[int, None]]]
    ):
        for pattern in shape_patterns:
            if cls._match_shape(shape, pattern):
                return True
        return False

    @classmethod
    def _shape_to_str(cls, shape):
        return "({})".format(", ".join(["N" if c is None else str(c) for c in shape]))

    @classmethod
    def _to_array_checked(cls, value, accepted_shapes, name, dtype=None) -> np.ndarray:
        assert len(accepted_shapes) > 0
        value = np.array(value)
        if dtype is not None:
            value = value.astype(dtype)
        if isinstance(accepted_shapes, tuple):
            accepted_shapes = [accepted_shapes]

        if not cls._match_shapes(value.shape, accepted_shapes):
            if len(accepted_shapes) == 1:
                shape_str = cls._shape_to_str(accepted_shapes[0])
            else:
                shape_str = " or ".join(
                    [
                        ", ".join(map(cls._shape_to_str, accepted_shapes[:-1])),
                        cls._shape_to_str(accepted_shapes[-1]),
                    ]
                )
            raise ValueError(
                f"Expected {name} of shape {shape_str}, but got array of shape {value.shape}."
            )
        return value



import numpy as np

from typing import List, Tuple, Union, Callable, Any

from armarx.arviz.elements import Box
from armarx.arviz.layer import Layer
from armarx.arviz.volumetric.size_models import SizeModel, FixedScaleSizeModel


def get_grid_coordinates(
    bounds: Union[np.ndarray, List[Tuple[float, float]]],
    num=None,
    flatten_dims=True,
    return_voxel_size=False,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Create coordinates in a 3D grid.

    The coordinates of a single point form the last axis.

    :param bounds:
        Limits in the form [(x0, x1), (y0, y1), (z0, z1)] (shape (d, 2))
        or [(x0, y0, z0), (x1, y1, z1)] (shape (2, d)).
    :param num:
        Number grid coordinates along each axis.
    :param flatten_dims:
        If true (default), the resulting array has shape (num^d, d),
        otherwise it has (num, num, num, d).
    :param return_voxel_size:
        If true, return the extents of a single voxel element, i.e.
        the 3D offset between two grid coordinates.

    :return: The grid coordinates with a the last axis representing a single point.
    """
    num = num or 16

    bounds = np.array(bounds)
    if bounds.ndim != 2:
        raise ValueError(f"Bounds must be 2D, but shape was {bounds.shape}.")
    if bounds.shape[-1] != 2 and bounds.shape[0] == 2:
        bounds = bounds.T

    xs = np.stack(np.meshgrid(*[np.linspace(a, b, num) for a, b in bounds]))
    xs = np.moveaxis(xs, 0, -1)
    voxel_size = xs[1, 1, 1] - xs[0, 0, 0]

    if flatten_dims:
        xs = xs.reshape(-1, len(bounds))

    if return_voxel_size:
        return xs, voxel_size
    else:
        return xs


class Volumetric:
    def __init__(
        self,
        bounds,
        num=None,
        alpha=None,
        cmap=None,
        size_model: SizeModel = None,
    ):

        self.bounds = bounds
        self.num = num or 16

        self.alpha = alpha

        if cmap is None:
            from matplotlib.cm import get_cmap

            cmap = get_cmap("viridis")
        elif isinstance(cmap, str):
            from matplotlib.cm import get_cmap

            cmap = get_cmap(cmap)
        self.cmap = cmap

        self.size_model = size_model or FixedScaleSizeModel(scale=1.0)
        self._size_model_context = SizeModel.Context()

    def draw_on(
        self,
        func: Callable[[Any], np.ndarray],
        layer: Layer,
        vlimits=None,
        id_prefix="",
    ):
        import matplotlib.pyplot as plt

        xs, voxel_size = get_grid_coordinates(
            self.bounds, self.num, return_voxel_size=True
        )
        ys = np.array(func(xs))

        if vlimits is None:
            vlimits = (ys.min(), ys.max())

        normalize = plt.Normalize(vmin=vlimits[0], vmax=vlimits[1])
        colors = self.cmap(normalize(ys))
        colors = (colors * 255).astype(np.int)
        if self.alpha is not None:
            colors[:, -1] = self.alpha

        self.size_model.set_value_limits(vlimits[0], vlimits[1])

        for i, (x, y, color) in enumerate(zip(xs, ys, colors)):
            self._size_model_context.i = i
            self._size_model_context.x = x
            self._size_model_context.y = y

            size = self.size_model.get_size(
                voxel_size=voxel_size, context=self._size_model_context
            )
            if size is None or min(size) <= 0:
                continue
            layer.add(Box(f"{id_prefix}{i:>04}", position=x, size=size, color=color))



def import_aron_slice():
    """
    Import and return the aron slice module.
    :return: The aron slice module.
    """

    def handle_error():
        import armarx
        from armarx import slice_loader

        slice_loader.load_armarx_slice("RobotAPI", "aron.ice")

    try:
        from armarx import aron

        # import armarx.aron
    except ImportError:
        handle_error()
        from armarx import aron

    return aron



import logging
import threading
import warnings

from abc import ABC
from abc import abstractmethod

from typing import Tuple
from typing import Any
from typing import Union

import numpy as np

from armarx_core.ice_manager import using_topic
from armarx_core.ice_manager import get_proxy
from armarx_core.ice_manager import register_object
from armarx_core.ice_manager import is_alive

from armarx_vision.image_provider import ImageProvider

from visionx import ImageProviderInterfacePrx
from visionx import ImageProcessorInterface

from armarx import MetaInfoSizeBase

from .shm_tools import path_to_shm
from .shm_tools import read_images_shm


logger = logging.getLogger(__name__)


class ImageProcessor(ImageProcessorInterface, ABC):
    """
    An abstract class  to process images

    .. highlight:: python
    .. code-block:: python

        class TestImageProcessor(ImageProcessor):

            def process_images(self, images, info):
                info.timestamp = time.time()
                return np.random.random(images.shape) * 128, info


    """

    def __init__(self, provider_name: str, num_result_images: int = None):
        super().__init__()
        self.provider_name = provider_name
        self.num_result_images = num_result_images

        self._thread = threading.Thread(target=self._process)
        self.image_available = False
        self.cv = threading.Condition()

        self.image_source = None
        self.result_image_provider = None

        self.shm_path = None

    def reportImageAvailable(self, provider_name, current=None):
        with self.cv:
            self.image_available = True
            self.cv.notify()

    def _get_images_and_info(self):
        if self.shm_path:
            info = MetaInfoSizeBase()  # self.image_source.getMetaInfo()
            images = read_images_shm(self.shm_path, self.data_dimensions)
        else:
            image_buffer, info = self.image_source.getImagesAndMetaInfo()
            images = np.frombuffer(image_buffer, dtype=np.uint8).reshape(
                self.data_dimensions
            )
        return images, info

    def _process(self):
        while is_alive():
            with self.cv:
                self.cv.wait_for(lambda: self.image_available)

                input_images, info = self._get_images_and_info()

                if hasattr(self, "process_image") and callable(self.process_image):
                    warnings.warn(
                        "Replaced with process_image(images, info)", DeprecationWarning
                    )
                    result = self.process_image(input_images)
                else:
                    result = self.process_images(input_images, info)

                if not result:
                    logger.warning("Unable to get images")
                    return
                elif isinstance(result, tuple):
                    result_images, info = result
                else:
                    result_images = result
                    info.timeProvided = 0

                self.result_image_provider.update_image(
                    result_images, info.timeProvided
                )

    @abstractmethod
    def process_images(
        self, images: np.ndarray, info: MetaInfoSizeBase
    ) -> Union[np.ndarray, Tuple[np.ndarray, MetaInfoSizeBase]]:
        """
        This function is called everytime a new image is available.
        Results are automatically published.

        :param images: the new images
        :param info: meta information about the image
        :returns: Either the result images only or a tuple containing the result image and the info
        """
        pass

    def register(self):
        warnings.warn("Replaced with on_connect", DeprecationWarning)
        self.on_connect()

    def shutdown(self):
        warnings.warn("Replaced with on_disconnect", DeprecationWarning)
        self.on_disconnect()

    def on_disconnect(self):
        self._image_listener_topic.unsubscribe(self._proxy)

    def on_connect(self):
        logger.debug("Registering image processor")
        proxy = register_object(self, self.__class__.__name__)
        self.image_source = get_proxy(ImageProviderInterfacePrx, self.provider_name)
        self.shm_path = path_to_shm(self.provider_name)

        number_of_images = self.image_source.getNumberImages()
        image_format = self.image_source.getImageFormat()
        d = image_format.dimension
        if self.num_result_images is None:
            self.num_result_images = self.image_source.getNumberImages()
        self.result_image_provider = ImageProvider(
            f"{self.__class__.__name__}Result",
            self.num_result_images,
            d.width,
            d.height,
        )
        self.result_image_provider.register()

        self.data_dimensions = (
            number_of_images,
            image_format.dimension.height,
            image_format.dimension.width,
            image_format.bytesPerPixel,
        )
        logger.debug("data dimensions %s", self.data_dimensions)

        self._thread.start()
        self._image_listener_topic = using_topic(
            proxy, f"{self.provider_name}.ImageListener"
        )



import dataclasses as dc
import typing as ty

from armarx_memory.aron.aron_dataclass import AronDataclass


@dc.dataclass
class Names(AronDataclass):

    spoken: ty.List[str] = dc.field(default_factory=list)
    recognized: ty.List[str] = dc.field(default_factory=list)



import enum
import numpy as np

from typing import List
from armarx.math.transform import Transform


from armarx import Vector3f
from armarx.viz.data import Element
from armarx.viz.data import InteractionFeedback
from armarx.viz.data import InteractionFeedbackType
from armarx.viz.data import CommitResult


class InteractionFeedbackType(enum.IntFlag):
    None_ = 0

    Select = 1
    Deselect = 2

    ContextMenuChosen = 3

    Transform = 4


class InteractionFeedback:
    def __init__(
        self,
        data: InteractionFeedback,
    ):
        self._data = data

    @property
    def type(self) -> InteractionFeedbackType:
        # Mask out all the flags in the higher bits
        ice_type = self._data.type & 0x7
        Types = InteractionFeedbackType
        IceTypes = InteractionFeedbackType

        type_dict = {
            IceTypes.NONE: Types.None_,
            IceTypes.SELECT: Types.Select,
            IceTypes.DESELECT: Types.Deselect,
            IceTypes.CONTEXT_MENU_CHOSEN: Types.ContextMenuChosen,
            IceTypes.TRANSFORM: Types.Transform,
        }

        try:
            return type_dict[ice_type]
        except KeyError:
            raise ValueError(f"Unexpected InteractionFeedbackType {ice_type}.")

    @property
    def is_transform_begin(self) -> bool:
        return self._data.type & InteractionFeedbackType.TRANSFORM_BEGIN_FLAG

    @property
    def is_transform_during(self) -> bool:
        return self._data.type & InteractionFeedbackType.TRANSFORM_DURING_FLAG

    @property
    def is_transform_end(self) -> bool:
        return self._data.type & InteractionFeedbackType.TRANSFORM_END_FLAG

    @property
    def layer(self) -> str:
        return self._data.layer

    @property
    def element(self) -> str:
        return self._data.element

    @property
    def revision(self) -> int:
        return self._data.revision

    @property
    def chosen_context_menu_entry(self) -> int:
        return self._data.chosenContextMenuEntry

    @property
    def transformation(self) -> Transform:

        from armarx.arviz.conv import GlobalPoseConv

        global_pose = self._data.transformation
        return self.global_pose_conv.from_ice(global_pose)

    @property
    def scale(self) -> np.ndarray:
        """
        :return: The scale as [x, y, z] array.
        """
        scale: Vector3f = self._data.scale
        return Vector3f(*scale)


class CommitResult:
    def __init__(
        self,
        data: CommitResult,
    ):
        self._data = data
        ice_interactions: List[InteractionFeedback] = self._data.interactions
        self.interactions = [InteractionFeedback(data) for data in ice_interactions]

    @property
    def revision(self) -> int:
        return self._data.revision



"""
This module provides functionality for receiving point clouds in ArmarX.

Classes:
- PointCloudReceiver: Can receive point clouds as numpy arrays.
"""

import logging
import threading

import numpy as np

from typing import Tuple

from armarx_core import ice_manager

from armarx_vision.pointclouds import dtype_from_point_type
from armarx_vision.pointclouds import PointCloudProcessorInterface
from armarx_vision.pointclouds import PointCloudProviderInterfacePrx
from armarx_vision.pointclouds import MetaPointCloudFormat


logger = logging.getLogger(__name__)


class PointCloudReceiver(PointCloudProcessorInterface):
    """
    A point cloud receiver connects to a PointCloudProvider and makes reads new point cloud data if available.
    """

    def __init__(
        self,
        name: str,
        source_provider_name: str = None,
        connect: bool = False,
        wait_for_provider=True,
    ):
        """
        Constructs a point cloud reciever.

        A point cloud receiver connects to a source point cloud provider.
        It grants easy access to the source provider's point clouds as numpy arrays.

        :param name: Name of the receiver component
        :param source_provider_name: Name of the source point cloud provider (implements PointCloudProviderInterface)
        :param connect: Indicates whether the constructor should automatically connect, i.e. call on_connect()
        """
        self.name = name
        self.proxy = None

        self.cv = threading.Condition()
        self.point_cloud_available = False

        # Source provider is set in on_connect()
        self.source_provider_name = source_provider_name
        self.source_provider_proxy = None
        self.source_provider_topic = None
        self.source_format = None

        self._wait_for_provider = wait_for_provider

        if connect:
            self.on_connect()

    def reportPointCloudAvailable(self, provider_name: str, current=None):
        with self.cv:
            self.point_cloud_available = True
            self.cv.notify()

    def wait_for_next_point_cloud(self) -> Tuple[np.array, MetaPointCloudFormat]:
        """
        Wait for the next point cloud from the source provider to arrive, then return it.

        This function blocks until a new point cloud is provided.

        :return: Tuple consisting of received point cloud data and format
        """
        with self.cv:
            self.cv.wait_for(lambda: self.point_cloud_available)
            return self.get_latest_point_cloud()

    def get_latest_point_cloud(self) -> Tuple[np.array, MetaPointCloudFormat]:
        """
        Get the latest point cloud without waiting.

        This function does not block but might return the same point cloud multiple times.

        :return: Tuple consisting of received point cloud data and format
        """
        raw_point_cloud, pc_format = self.source_provider_proxy.getPointCloud()
        # FIXME: Why can pc_format be not set here? It is an output parameter that should always be set.
        if pc_format is None:
            pc_format = self.source_format

        point_dtype = dtype_from_point_type(pc_format.type)
        point_cloud = np.frombuffer(raw_point_cloud, dtype=point_dtype)

        return point_cloud, pc_format

    def on_disconnect(self):
        """
        Call this function after you have finished receiving point clouds.
        """
        self.source_provider_topic.unsubscribe(self.proxy)

    def on_connect(self):
        """
        This function starts the connection with the source provider.

        After calling this function, wait_for_next_point_cloud() and get_latest_point_cloud() can be called.
        """
        logger.debug("Registering point cloud processor")
        self.proxy = ice_manager.register_object(self, self.name)
        if self._wait_for_provider:
            self.source_provider_proxy = ice_manager.wait_for_proxy(
                PointCloudProviderInterfacePrx, self.source_provider_name
            )
        else:
            self.source_provider_proxy = ice_manager.get_proxy(
                PointCloudProviderInterfacePrx, self.source_provider_name
            )
        self.source_format = self.source_provider_proxy.getPointCloudFormat()

        self.source_provider_topic = ice_manager.using_topic(
            self.proxy, f"{self.source_provider_name}.PointCloudListener"
        )



import enum
import numpy as np

from typing import Iterable, Union, List

from armarx import slice_loader

slice_loader.load_armarx_slice("RobotAPI", "ArViz.ice")

from armarx.viz.data import Element

from armarx.arviz import conversions

from armarx.math.transform import Transform
from armarx import Vector3f


class ElementFlags(enum.IntFlag):
    NONE = 0
    OVERRIDE_MATERIAL = 1
    HIDDEN = 2


class Element:
    def __init__(
        self,
        ice_data_cls,
        id,
        pose=None,
        position=None,
        orientation=None,
        color=None,
        scale=1.0,
    ):
        self.ice_data_cls = ice_data_cls
        self.id: str = str(id)

        self.pose = np.eye(4)
        if pose is not None:
            self.pose = pose
        if position is not None:
            self.position = position
        if orientation is not None:
            self.ori_mat = orientation

        self.color = color if color is not None else (100, 100, 100, 255)

        self.scale = scale
        self.flags: ElementFlags = ElementFlags.NONE

        from armarx.viz.data import InteractionDescription

        self._interaction = InteractionDescription()

    @property
    def pose(self) -> np.ndarray:
        """The pose as 4x4 transformation matrix."""
        return self._pose

    @pose.setter
    def pose(self, value):
        if isinstance(value, Transform):
            value = value.transform

        value = self._to_array_checked(value, (4, 4), "pose matrix", dtype=np.float)
        self._pose = value

    @property
    def position(self) -> np.ndarray:
        return self._pose[:3, 3]

    @position.setter
    def position(self, value):
        value = self._to_array_checked(
            value, [(3,), (3, 1)], "position vector", dtype=np.float
        )
        self._pose[:3, 3] = value

    @property
    def orientation(self):
        return self.ori_mat

    @orientation.setter
    def orientation(self, value):
        value = np.array(value)
        if value.shape == (4,):
            self.ori_quat = value
        elif value.shape == (3, 3):
            self.ori_mat = value
        else:
            raise ValueError("Invalid orientation value: {}".format(value))

    @property
    def ori_mat(self) -> np.ndarray:
        """The orientation as 3x3 rotation matrix."""
        return self._pose[:3, :3]

    @ori_mat.setter
    def ori_mat(self, value):
        """Set the orientation as 3x3 rotation matrix."""
        value = self._to_array_checked(
            value, (3, 3), "orientation matrix", dtype=np.float
        )
        self._pose[:3, :3] = value

    @property
    def ori_quat(self):
        """The orientation as [w, x, y, z] quaternion."""
        import transforms3d as tf3d

        try:
            return tf3d.quaternions.mat2quat(self.ori_mat)
        except np.linalg.LinAlgError as e:
            print("Rotation matrix to quaternion: \n{}".format(self.ori_mat))
            raise

    @ori_quat.setter
    def ori_quat(self, value):
        import transforms3d as tf3d

        try:
            self.ori_mat = tf3d.quaternions.quat2mat(value)
        except np.linalg.LinAlgError as e:
            print("Quaternion to rotation matrix: \n{}".format(value))
            raise

    @property
    def color(self) -> np.ndarray:
        return self._color

    @color.setter
    def color(self, value):
        value = self._to_array_checked(value, [(3,), (4,)], "color")
        if value.shape == (3,):
            self._color = np.concatenate([value, [255]])
        else:
            self._color = value

    @property
    def scale(self) -> np.ndarray:
        """
        :return: A scaling vector [x, y, z] of shape (3,).
        """
        return self._scale

    @scale.setter
    def scale(self, value: Union[float, np.ndarray]):
        if isinstance(value, (float, int)):
            self._scale = np.array([value, value, value], dtype=float)
        else:
            value = np.array(value)
            if not value.shape == (3,):
                raise ValueError(
                    f"Expected scale to be scalar or an array of shape (3,), "
                    f"but got shape ({value.shape})."
                )
            self._scale = value

    # Interaction

    def enable_interaction(
        self,
        selection: bool = False,
        context_menu_options: List[str] = None,
        translation: str = None,
        rotation: str = None,
        scaling: str = None,
        transform: bool = False,
        hide_during_transform: bool = False,
    ) -> "Element":
        """
        Enable one or more interaction features.

        :param selection:
            Enable selection / deselection. Implied by most other features.
        :param context_menu_options:
            Enable a context menu showing the given options.
        :param translation:
            Enable translation along a set of axis, specified by a subset
            of "xyzl" (with l for local axes).
        :param rotation:
            Enable rotation around a set of axis, specified by a subset
            of "xyzl" (with l for local axes).
        :param scaling:
            Enable scaling along a set of axis, specified by a subset
            of "xyzl" (with l for local axes).
        :param transform:
            Implies translation = "xyz" and rotation = "xyz".
        :param hide_during_transform:
            Hide the original element while a ghost is shown during interaction.
            If False, both the original element and the ghost are visible.
        :return: None
        """
        from armarx.viz.data import InteractionEnableFlags as Flags

        flags: int = self._interaction.enableFlags

        if context_menu_options:
            flags |= Flags.CONTEXT_MENU
            selection = True

        if transform:
            translation = "xyz"
            rotation = "xyz"

        if translation:
            for char, flag in [
                ("x", Flags.TRANSLATION_X),
                ("y", Flags.TRANSLATION_Y),
                ("z", Flags.TRANSLATION_Z),
                ("l", Flags.TRANSLATION_LOCAL),
            ]:
                if char in translation:
                    flags |= flag
            selection = True

        if rotation:
            for char, flag in [
                ("x", Flags.ROTATION_X),
                ("y", Flags.ROTATION_Y),
                ("z", Flags.ROTATION_Z),
                ("l", Flags.ROTATION_LOCAL),
            ]:
                if char in rotation:
                    flags |= flag
            selection = True

        if scaling:
            for char, flag in [
                ("x", Flags.SCALING_X),
                ("y", Flags.SCALING_Y),
                ("z", Flags.SCALING_Z),
                ("l", Flags.SCALING_LOCAL),
            ]:
                if char in scaling:
                    flags |= flag
            selection = True

        if selection:
            flags |= Flags.SELECT

        if hide_during_transform:
            flags |= Flags.TRANSFORM_HIDE

        self._interaction.enableFlags = flags
        self._interaction.contextMenuOptions = context_menu_options or []

        return self

    # Behind the scenes

    def get_ice_data(self):
        """Get the Ice data for committing."""
        ice_data = self.ice_data_cls(id=self.id, interaction=self._interaction)
        self._update_ice_data(ice_data)
        return ice_data

    def _update_ice_data(self, ice_data):
        """
        Update the given element data with self's values.
        :param data: armarx.viz.data.Element
        """
        p = ice_data.pose
        p.x, p.y, p.z = self.position
        p.qw, p.qx, p.qy, p.qz = map(float, self.ori_quat)

        c = ice_data.color
        c.r, c.g, c.b, c.a = map(int, self.color)
        if isinstance(self.scale, float):
            scale = np.array([self.scale, self.scale, self.scale])
        else:
            assert self.scale.shape == (
                3,
            ), f"Expected scale of shape {(3,)}, but got {self.scale.shape}."
            scale = self.scale
        ice_data.scale = Vector3f(*scale)

        ice_data.flags = int(self.flags)

    @classmethod
    def _match_shape(
        cls, shape: Iterable[int], shape_pattern: Iterable[Union[int, None]]
    ):
        """
        Indicate whether `shape` matches `accepted_shape`.
        :param shape: An array shape, i.e. tuple of ints.
        :param shape_pattern:
            A shape pattern, i.e. tuple of ints or None, where None matches any size.
        :return: True if `shape` matches `accepted_shape`.
        """
        return np.all(
            np.logical_or(
                np.array(shape) == shape_pattern,
                np.isnan(np.array(shape_pattern, dtype=np.float)),
            )
        )

    @classmethod
    def _match_shapes(
        cls, shape: Iterable[int], shape_patterns: List[Iterable[Union[int, None]]]
    ):
        for pattern in shape_patterns:
            if cls._match_shape(shape, pattern):
                return True
        return False

    @classmethod
    def _shape_to_str(cls, shape):
        return "({})".format(", ".join(["N" if c is None else str(c) for c in shape]))

    @classmethod
    def _to_array_checked(cls, value, accepted_shapes, name, dtype=None) -> np.ndarray:
        assert len(accepted_shapes) > 0
        value = np.array(value)
        if dtype is not None:
            value = value.astype(dtype)
        if isinstance(accepted_shapes, tuple):
            accepted_shapes = [accepted_shapes]

        if not cls._match_shapes(value.shape, accepted_shapes):
            if len(accepted_shapes) == 1:
                shape_str = cls._shape_to_str(accepted_shapes[0])
            else:
                shape_str = " or ".join(
                    [
                        ", ".join(map(cls._shape_to_str, accepted_shapes[:-1])),
                        cls._shape_to_str(accepted_shapes[-1]),
                    ]
                )
            raise ValueError(
                "Expected {} of shape {}, but got array of shape {}.".format(
                    name, shape_str, value.shape
                )
            )
        return value



from datetime import datetime

import numpy as np

from armarx import VariantBase
from armarx import TimedVariantBase

from armarx import StringVariantData
from armarx import FloatVariantData
from armarx import BoolVariantData
from armarx import IntVariantData


def hash_type_name(type_id: str) -> int:
    """
    converts an ice id to a variant's type id

    ..see:: c++ implementation in ArmarXCore/observers/variants/Variants.cpp::hashTypeName()

    The implementation uses a normal int, thus the value can be larger than 2 ** 31 - 1

    :param type_id:: the type id
    :returns: the hash value of type_id
    """
    prev_error_level = np.geterr()
    np.seterr(over="ignore")
    hash_value = np.int32(0)
    for ch in type_id:
        hash_value = (
            (np.int32(hash_value) << np.int32(5)) + np.int32(hash_value)
        ) ^ np.int32(ord(ch))
    np.seterr(over=prev_error_level["over"])
    return hash_value


def convert_to_variant_data(data):
    """
    Wraps python data into an ArmarX variant
    """
    if isinstance(data, str):
        return StringVariantData(data)
    elif isinstance(data, float):
        return FloatVariantData(data)
    elif isinstance(data, bool):
        return BoolVariantData(data)
    elif isinstance(data, int):
        return IntVariantData(data)
    return data


class TimedVariant(TimedVariantBase):
    def __init__(self, data=None, _typeId=-1, _timestamp=0):
        super().__init__(data, _typeId=-1, _timestamp=0)
        if self._typeId == -1 and hasattr(data, "ice_id"):
            self._typeId = hash_type_name(self.data.ice_id())
        if _timestamp == 0:
            self._timestamp = datetime.now().timestamp()


class Variant(VariantBase):
    def __init__(self, data=None, _typeId=-1):
        super().__init__(data, _typeId)
        if self._typeId == -1 and hasattr(data, "ice_id"):
            self._typeId = hash_type_name(self.data.ice_id())



from typing import Dict, List, Optional
import logging

import armarx
from armarx_core import slice_loader
from armarx_core import ice_manager

slice_loader.load_armarx_slice("RobotAPI", "armem/server/MemoryInterface.ice")
slice_loader.load_armarx_slice("RobotAPI", "armem/mns/MemoryNameSystemInterface.ice")

from armarx import armem


from armarx_memory.core import MemoryID, error as armem_error
from armarx_memory.client.Reader import Reader
from armarx_memory.client.Writer import Writer


class ServerProxies:
    def __init__(
        self,
        reading: Optional[armem.server.ReadingMemoryInterface] = None,
        writing: Optional[armem.server.WritingMemoryInterface] = None,
    ):
        self.reading = reading
        self.writing = writing

    @classmethod
    def from_ice(cls, server: armem.mns.dto.MemoryServerInterfaces):
        return cls(reading=server.reading, writing=server.writing)


class MemoryNameSystem:

    cls_logger = logging.getLogger(__file__)

    MemoryNameSystemPrx = "armarx.armem.mns.MemoryNameSystemInterfacePrx"
    MemoryServerPrx = "armarx.armem.server.MemoryInterfacePrx"

    @classmethod
    def get_mns(cls, mns_name="MemoryNameSystem", **kwargs) -> "MemoryNameSystem":
        import Ice

        try:
            mns_proxy = ice_manager.get_proxy(
                armarx.armem.mns.MemoryNameSystemInterfacePrx, mns_name
            )
            return MemoryNameSystem(mns_proxy, **kwargs)

        except Ice.NotRegisteredException as e:
            cls.cls_logger.error(e)
            raise armem_error.ArMemError(
                f"Memory Name System '{MemoryNameSystem}' is not registered."
            )

    @classmethod
    def wait_for_mns(
        cls, mns_name="MemoryNameSystem", logger=None, **kwargs
    ) -> "MemoryNameSystem":
        if logger is not None:
            logger.info("Wait for Memory Name System ... ")

        mns_proxy = ice_manager.wait_for_proxy(
            armem.mns.MemoryNameSystemInterfacePrx, mns_name, timeout=0
        )

        if logger is not None:
            logger.info("Done.")

        return MemoryNameSystem(mns_proxy, **kwargs)

    def __init__(
        self,
        mns: Optional[MemoryNameSystemPrx],
        logger=None,
    ):

        self.mns = mns
        self.servers: Dict[str, ServerProxies] = {}

        self.logger = logger or self.cls_logger

    # Server Resolution

    def update(self):
        import Ice

        result: "armem.data.GetAllRegisteredMemoriesResult"
        try:
            result = self.mns.getAllRegisteredServers()
        except Ice.NotRegisteredException as e:
            raise armem_error.ArMemError(e)
        if result.success:
            # Do some implicit type check
            self.servers = {
                name: ServerProxies.from_ice(server)
                for name, server in result.servers.items()
            }
        else:
            raise armem_error.ArMemError(
                f"MemoryNameSystem query failed: {result.errorMessage}"
            )

    def resolve_server(
        self,
        memory_id: MemoryID,
    ) -> ServerProxies:

        server = self.servers.get(memory_id.memory_name, None)

        if server is None:
            self.update()
            server = self.servers.get(memory_id.memory_name, None)
            if server is None:
                raise armem_error.CouldNotResolveMemoryServer(memory_id)

        assert server is not None
        return server

    def wait_for_server(self, memory_id: MemoryID) -> ServerProxies:

        server = self.servers.get(memory_id.memory_name, None)
        if server is None:
            inputs = armem.mns.dto.WaitForServerInput(
                name=memory_id.memory_name,
            )

            self.logger.info(f"Waiting for memory server {memory_id} ...")
            result: armem.mns.dto.WaitForServerResult = self.mns.waitForServer(inputs)
            self.logger.info(f"Resolved memory server {memory_id}.")
            if result.success:
                if result.server.reading or result.server.writing:
                    server = ServerProxies.from_ice(result.server)
                else:
                    raise armem_error.CouldNotResolveMemoryServer(
                        memory_id, f"Returned proxy is null: {result.proxy}"
                    )
            else:
                raise armem_error.CouldNotResolveMemoryServer(
                    memory_id, result.errorMessage
                )

        assert server is not None
        return server

    def get_reader(self, memory_id: MemoryID) -> Reader:
        return Reader(self.resolve_server(memory_id).reading)

    def wait_for_reader(self, memory_id: MemoryID) -> Reader:
        return Reader(self.wait_for_server(memory_id).reading)

    def get_all_readers(self, update=True) -> Dict[str, Reader]:
        return self._get_all_clients(Reader, update)

    def get_writer(self, memory_id: MemoryID) -> Writer:
        return Writer(self.resolve_server(memory_id).writing)

    def wait_for_writer(self, memory_id: MemoryID) -> Writer:
        return Writer(self.wait_for_server(memory_id).writing)

    def get_all_writers(self, update=True) -> Dict[str, Writer]:
        return self._get_all_clients(Writer, update)

    # System-wide queries

    def resolve_entity_instance(
        self,
        id: MemoryID,
    ) -> Optional["armem.data.EntityInstance"]:
        instances = self.resolve_entity_instances([id])
        if len(instances) > 0:
            return instances[id]
        else:
            return None

    def resolve_entity_snapshots(
        self,
        ids: List[MemoryID],
    ) -> Dict[MemoryID, "armem.data.EntitySnapshot"]:
        errors = ""
        error_counter = 0

        ids_per_memory = dict()
        for id in ids:
            if id.memory_name in ids_per_memory:
                ids_per_memory[id.memory_name].append(id)
            else:
                ids_per_memory[id.memory_name] = [id]

        result = dict()
        for memory_name, ids in ids_per_memory.items():
            reader = self.get_reader(memory_id=MemoryID(memory_name))
            try:
                snapshots = reader.query_snapshots(ids)
                result = {**result, **snapshots}  # Merge dicts
            except RuntimeError as e:
                error_counter += 1
                errors += f"\n#{error_counter}\n"
                errors += f"Failed to retrieve IDs {ids} from query result: \n{e}"

        if errors:
            self.logger.info(
                f"{self.__class__.__name__}.{self.resolve_entity_instance.__name__}:"
                f"The following errors may affect your result: \n{errors}\n\n"
                + "When resolving entity snapshots: \n- {}".format(
                    "\n- ".join(map(str, ids))
                )
            )

        return result

    # ToDo: System-wide commits

    def _get_all_clients(self, client_cls, update: bool):
        if update:
            self.update()
        return {name: client_cls(server) for name, server in self.servers}

    def __bool__(self):
        return bool(self.mns)

    @classmethod
    def get_server_by_proxy_name(
        cls,
        proxy_name: str,
    ) -> "armarx.armem.server.MemoryInterfacePrx":
        """
        Get a memory server proxy by its Ice object name.
        :param proxy_name: The ice object name.
        :return: The server proxy, if it exists.
        :throw: Ice.NotRegisteredException If the server does not exist.
        """
        import Ice

        try:
            return ice_manager.get_proxy(armem.server.MemoryInterfacePrx, proxy_name)

        except Ice.NotRegisteredException as e:
            cls.cls_logger.error(e)
            return None



import time


class Metronome:
    """
    A cycle clock allowing you to run a while-loop with an (approximately)
    fixed frequency.

    Example for 10 Hz:

    cycle = CycleClock(frequency_hertz=10)  # alternative: duration_seconds=0.01
    while True:
        ...  # Do stuff.

        cycle.wait_for_next_cycle()
    """

    def __init__(
        self,
        duration_seconds: float = None,
        frequency_hertz: float = None,
    ):
        self.duration_seconds: float
        if duration_seconds is not None:
            self.duration_seconds = duration_seconds

        elif frequency_hertz is not None:
            self.duration_seconds = 1 / frequency_hertz

        else:
            raise ValueError(
                "Either duration_seconds or frequency_hertz must be specified."
            )

        self.start: float = 0.0
        self.reset()

    def reset(self):
        self.start = time.time()

    @property
    def remaining_seconds(self) -> float:
        return self.duration_seconds - (time.time() - self.start)

    def wait_for_next_tick(self):
        remaining = self.remaining_seconds
        if remaining > 0:
            time.sleep(remaining)
        self.reset()



from abc import ABC
from abc import abstractmethod

from typing import List
from typing import Optional


from armarx_memory.core import MemoryID
from armarx_memory.client import MemoryNameSystem
from armarx_memory.client import Commit
from armarx_memory.client import Reader
from armarx_memory.client import Writer


class GenericDataClass:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_aron(self) -> "armarx.aron.data.dto.GenericData":
        from armarx_memory.aron.conversion import to_aron

        return to_aron(self.__dict__)

    @classmethod
    def from_aron(cls, dto: "armarx.aron.data.dto.GenericData"):
        from armarx_memory.aron.conversion import from_aron

        return cls(**from_aron(dto))


class BaseClient(ABC):
    @property
    @abstractmethod
    def core_segment_id(self):
        pass

    @property
    @abstractmethod
    def default_entity_name(self):
        pass

    def make_entity_name(self, provider_name: str, entity_name: str = None):
        if not entity_name:
            entity_name = self.default_entity_name
        return self.core_segment_id.with_provider_segment_name(
            provider_name
        ).with_entity_name(entity_name)


class BaseReader(Reader, BaseClient, ABC):
    def fetch_latest_instance(self, updated_ids: Optional[List[MemoryID]] = None):
        """
        Query the latest snapshot of the given updated IDs and
        return its first instance.
        """
        if updated_ids is None:
            memory = self.query_latest(self.core_segment_id)

            latest_snapshot = None

            core_seg = memory.coreSegments[self.core_segment_id.core_segment_name]
            for prov_seg in core_seg.providerSegments.values():
                for entity in prov_seg.entities.values():
                    for snapshot in entity.history.values():
                        if latest_snapshot is None:
                            latest_snapshot = snapshot
                        elif (
                            latest_snapshot.id.timestamp.timeSinceEpoch.microSeconds
                            < snapshot.id.timestamp.timeSinceEpoch.microSeconds
                        ):
                            latest_snapshot = snapshot
        else:
            for up_id in updated_ids:
                assert self.core_segment_id.contains(up_id)

            latest_snapshot_id = max(updated_ids, key=lambda i: i.timestamp_usec)
            latest_snapshot = self.query_snapshot(latest_snapshot_id)

        if not latest_snapshot:
            return None

        latest_instance = latest_snapshot.instances[0]
        return latest_instance

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem = None, wait=True):
        if not mns:
            mns = MemoryNameSystem.get_mns()
        if wait:
            return cls(mns.wait_for_server(cls.core_segment_id).reading)
        else:
            return cls(mns.resolve_server(cls.core_segment_id).reading)


class BaseWriter(BaseClient, ABC):
    def __init__(self, writer: Writer):
        super().__init__()
        self.writer = writer

    def commit(self, entity_id: MemoryID, time_created_usec=None, **kwargs):
        from armarx_memory.aron.conversion import to_aron

        commit = Commit()
        commit.add(
            entity_id=entity_id,
            time_created_usec=time_created_usec,
            instances_data=[to_aron(kwargs)],
        )
        return self.writer.commit(commit.to_ice())

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem = None, wait=True):
        if not mns:
            mns = MemoryNameSystem.get_mns()
        if wait:
            return cls(mns.wait_for_server(cls.core_segment_id).writing)
        else:
            return cls(mns.resolve_server(cls.core_segment_id).writing)



from armarx_memory.core.MemoryID import MemoryID


class ArMemError(Exception):
    """Base class for exceptions thrown by armarx.armem."""

    pass


class CouldNotResolveMemoryServer(ArMemError):
    def __init__(self, memory_id: MemoryID, msg: str = ""):
        super().__init__(
            f"Could not resolve the memory name {memory_id}."
            f"\nMemory server for {memory_id} is not registered."
            + ("\n" + msg if msg else "")
        )



import Ice

from armarx_core.ice_manager import ice_communicator
from armarx_core.slice_loader import load_armarx_slice

load_armarx_slice("RobotAPI", "observers/KinematicUnitObserverInterface.ice")

from armarx import DatafieldRefBase
from armarx import ChannelRefBase

from armarx_core.variants import Variant
from armarx_core.variants import TimedVariant


class VariantFactory(Ice.ObjectFactory):
    def create(self, t):
        return Variant()


class TimedVariantFactory(Ice.ObjectFactory):
    def create(self, t):
        return TimedVariant()


class DatafieldRef(DatafieldRefBase):
    pass


class DatafieldRefFactory(Ice.ObjectFactory):
    def create(self, t):
        return DatafieldRef()


class ChannelRef(ChannelRefBase):
    pass


class ChannelRefFactory(Ice.ObjectFactory):
    def create(self, t):
        return ChannelRef()


load_armarx_slice("RobotAPI", "core/FramedPoseBase.ice")
from armarx import PoseBase
from armarx import FramedPositionBase
from armarx import FramedPoseBase
from armarx import Vector3Base
from armarx import QuaternionBase

load_armarx_slice("ArmarXCore", "observers/Timestamp.ice")
from armarx import TimestampBase


class Vector3(Vector3Base):
    pass


class Vector3Factory(Ice.ObjectFactory):
    def create(self, t):
        return Vector3()


class Quaternion(QuaternionBase):
    pass


class QuaternionFactory(Ice.ObjectFactory):
    def create(self, t):
        return Quaternion()


class Pose(PoseBase):
    pass


class PoseFactory(Ice.ObjectFactory):
    def create(self, t):
        return Pose()


class FramedPosition(FramedPositionBase):
    pass


class FramedPositionFactory(Ice.ObjectFactory):
    def create(self, t):
        return FramedPosition()


class FramedPose(FramedPoseBase):
    pass


class FramedPoseFactory(Ice.ObjectFactory):
    def create(self, t):
        return FramedPose()


class Timestamp(TimestampBase):
    pass


class TimestampFactory(Ice.ObjectFactory):
    def create(self, t):
        return Timestamp()


def register():
    ice_communicator.addObjectFactory(VariantFactory(), Variant.ice_staticId())
    ice_communicator.addObjectFactory(
        TimedVariantFactory(), TimedVariant.ice_staticId()
    )
    ice_communicator.addObjectFactory(
        DatafieldRefFactory(), DatafieldRef.ice_staticId()
    )
    ice_communicator.addObjectFactory(ChannelRefFactory(), ChannelRef.ice_staticId())

    ice_communicator.addObjectFactory(TimestampFactory(), Timestamp.ice_staticId())
    ice_communicator.addObjectFactory(Vector3Factory(), Vector3.ice_staticId())
    ice_communicator.addObjectFactory(QuaternionFactory(), Quaternion.ice_staticId())
    ice_communicator.addObjectFactory(PoseFactory(), Pose.ice_staticId())
    ice_communicator.addObjectFactory(
        FramedPositionFactory(), FramedPosition.ice_staticId()
    )
    ice_communicator.addObjectFactory(FramedPoseFactory(), FramedPose.ice_staticId())



from typing import List
from typing import Optional

import numpy as np

from armarx_memory.core import MemoryID
from armarx_memory.client import MemoryNameSystem, Commit, Reader, Writer


class ObjectInstance(object):
    def __init__(self, classID: MemoryID, sourceID: MemoryID, pose: np.ndarray):
        self.classID = classID
        self.sourceID = sourceID
        self.pose = pose

    def to_aron(self) -> "armarx.aron.data.dto.GenericData":
        from armarx_memory.aron.conversion import to_aron

        dto = to_aron(
            {
                "classID": self.classID,
                "sourceID": self.sourceID,
                "pose": self.pose,
            }
        )
        return dto

    @classmethod
    def from_aron(cls, dto: "armarx.aron.data.dto.GenericData"):
        from armarx_memory.aron.conversion import from_aron

        d = from_aron(dto)
        return cls(**d)


class ObjectInstanceClientBase:

    core_segment_id = MemoryID("Object", "Instance")

    def __init__(self):
        pass

    def make_entity_name(
        self, provider_name: str, entity_name: str = "object_instance"
    ):
        return self.core_segment_id.with_provider_segment_name(
            provider_name
        ).with_entity_name(entity_name)


class ObjectInstanceWriter(ObjectInstanceClientBase):
    def __init__(self, writer: Writer):
        super().__init__()
        self.writer = writer

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem, wait=True) -> "ObjectInstanceWriter":
        return cls(
            mns.wait_for_writer(cls.core_segment_id)
            if wait
            else mns.get_writer(cls.core_segment_id)
        )

    def commit(
        self,
        entity_id: MemoryID,
        classID: MemoryID,
        sourceID: MemoryID,
        pose: np.ndarray,
        time_created_usec=None,
        **kwargs
    ):
        object_instance = ObjectInstance(classID=classID, sourceID=sourceID, pose=pose)
        commit = Commit()
        commit.add(
            entity_id=entity_id,
            time_created_usec=time_created_usec,
            instances_data=[object_instance.to_aron()],
            **kwargs
        )
        return self.writer.commit(commit)


class ObjectInstanceReader(ObjectInstanceClientBase):
    def __init__(self, reader: Reader):
        super().__init__()
        self.reader = reader

    def fetch_all_instances(self):
        entities = []
        memory = self.reader.query_latest(self.core_segment_id)

        core_seg = memory.coreSegments[self.core_segment_id.core_segment_name]
        for prov_seg in core_seg.providerSegments.values():
            for entity in prov_seg.entities.values():
                latest_snapshot = None
                for snapshot in entity.history.values():
                    if latest_snapshot is None:
                        latest_snapshot = snapshot
                    elif (
                        latest_snapshot.id.timestamp.timeSinceEpoch.microSeconds
                        < snapshot.id.timestamp.timeSinceEpoch.microSeconds
                    ):
                        latest_snapshot = snapshot

                if latest_snapshot:
                    entities.append(latest_snapshot.instances[0])

        return entities

    def fetch_latest_instance(self, updated_ids: Optional[List[MemoryID]] = None):
        """
        Query the latest snapshot of the given updated IDs and
        return its first instance.
        """
        if updated_ids is None:
            memory = self.reader.query_latest(self.core_segment_id)

            latest_snapshot = None

            core_seg = memory.coreSegments[self.core_segment_id.core_segment_name]
            for prov_seg in core_seg.providerSegments.values():
                for entity in prov_seg.entities.values():
                    for snapshot in entity.history.values():
                        if latest_snapshot is None:
                            latest_snapshot = snapshot
                        elif (
                            latest_snapshot.id.timestamp.timeSinceEpoch.microSeconds
                            < snapshot.id.timestamp.timeSinceEpoch.microSeconds
                        ):
                            latest_snapshot = snapshot
        else:
            for up_id in updated_ids:
                assert self.core_segment_id.contains(up_id)

            latest_snapshot_id = max(updated_ids, key=lambda i: i.timestamp_usec)
            latest_snapshot = self.reader.query_snapshot(latest_snapshot_id)

        if not latest_snapshot:
            return None

        latest_instance = latest_snapshot.instances[0]
        return latest_instance

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem, wait=True) -> "ObjectInstanceReader":
        return cls(
            mns.wait_for_reader(cls.core_segment_id)
            if wait
            else mns.get_reader(cls.core_segment_id)
        )



import numpy as np

from typing import List, Optional, Union

import armarx
from armarx import slice_loader
from armarx_memory.ice_conv.ice_converter import IceConverter

SLICE_INCLUDE = ("ArmarXCore", "core/BasicVectorTypes.ice")


class Vector2fConv(IceConverter):
    def __init__(self):
        super().__init__()
        self.set_handler_from_ice(list, self._from_ice)
        self.set_handler_to_ice(np.ndarray, self._to_ice)

    def _import_dto(cls):
        slice_loader.load_armarx_slice(*SLICE_INCLUDE)
        return armarx.Vector2f

    def _from_ice(
        self,
        dto: Union["armarx.Vector2f", List["armarx.Vector2f"]],
        scaling: Optional[float] = None,
    ) -> np.ndarray:

        if isinstance(dto, list):
            points = dto
            bo = np.array([(p.e0, p.e1) for p in points])
            assert bo.shape == (
                len(points),
                2,
            ), "Shape should be {}, but was {}.\n\tPoints: {}".format(
                (len(points), 2), bo.shape, points
            )
        else:
            point = dto
            bo = np.array((point.e0, point.e1))

        bo = scale(bo, scaling)
        return bo

    def _to_ice(
        self,
        bo: np.ndarray,
        scaling: Optional[float] = None,
    ) -> "armarx.Vector2f":
        Vector2f = self.get_dto()

        bo = np.array(bo)
        bo = scale(bo, scaling)

        if bo.ndim == 1:
            x, y = bo
            return Vector2f(x, y)
        else:
            return [Vector2f(x, y) for x, y in bo]


class Vector3fConv(IceConverter):
    def __init__(self):
        super().__init__()
        self.set_handler_from_ice(list, self._from_ice)
        self.set_handler_to_ice(np.ndarray, self._to_ice)

    def _import_dto(cls):
        slice_loader.load_armarx_slice(*SLICE_INCLUDE)
        return armarx.Vector3f

    def _from_ice(
        self,
        dto_points: Union["armarx.Vector3f", List["armarx.Vector3f"]],
        scaling: Optional[float] = None,
    ) -> np.ndarray:

        try:
            iter(dto_points)
        except TypeError:
            p = dto_points
            bo = np.array([p.e0, p.e1, p.e2])

        else:
            bo = np.array([(p.e0, p.e1, p.e2) for p in dto_points])
            assert bo.shape == (
                len(dto_points),
                3,
            ), "Shape should be {}, but was {}.\n\tPoints: {}".format(
                (len(dto_points), 3), bo.shape, dto_points
            )

        bo = scale(bo, scaling)
        return bo

    def _to_ice(
        self,
        bo: np.ndarray,
        scaling: Optional[float] = None,
    ) -> "armarx.Vector3f":
        Vector3f = self.get_dto()

        bo = scale(bo, scaling)

        if bo.ndim == 1:
            x, y, z = bo
            return Vector3f(x, y, z)
        else:
            return [Vector3f(x, y, z) for x, y, z in bo]


def scale(points: np.ndarray, scaling: Optional[float] = None) -> np.ndarray:
    points = np.array(points)
    return points if scaling is None else points * scaling



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['armarx',
 'armarx.arviz',
 'armarx.arviz.elements',
 'armarx.arviz.grids',
 'armarx.arviz.volumetric',
 'armarx.ice_conv',
 'armarx.ice_conv.armarx_core',
 'armarx.ice_conv.robot_api',
 'armarx.math',
 'armarx.remote_gui',
 'armarx.remote_gui.widgets',
 'armarx.tools',
 'armarx_core',
 'armarx_memory',
 'armarx_memory.aron',
 'armarx_memory.aron.aron_ice_types',
 'armarx_memory.aron.common',
 'armarx_memory.aron.conversion',
 'armarx_memory.client',
 'armarx_memory.client.detail',
 'armarx_memory.core',
 'armarx_memory.ice_conv',
 'armarx_memory.ice_conv.ArmarXCore',
 'armarx_memory.ice_conv.RobotAPI',
 'armarx_memory.segments',
 'armarx_memory.segments.human',
 'armarx_objects',
 'armarx_robots',
 'armarx_vision',
 'visionx']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.8.0', 'numpy>=1.19.5', 'transforms3d>=0.4.1', 'zeroc-ice==3.7.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.8']}

setup_kwargs = {
    'name': 'armarx',
    'version': '0.23.2',
    'description': 'A Python Toolbox for ArmarX',
    'long_description': "# Python ArmarX - A Python toolbox for ArmarX\n\n```\nfrom armarx import 🤖  as ❤ \n```\n\nThis package provides Python 3 bindings for ArmarX.\n\nIn addition, the package also includes some helper functions, such as\npublishing or subscribing to images.\n\n\n## Installation\n\n`pip install --upgrade --extra-index-url https://pypi.humanoids.kit.edu/ armarx`\n\n## Using the bindings\n\n### Connecting to an existing proxy\n\nFor proxies defined in a project's `Variants-*.xml` it is possible to import\nthe interface directly. \n\n```python\nfrom armarx import PlatformUnitInterfacePrx\nplatform_unit = PlatformUnitInterfacePrx.get_proxy('Armar6PlatformUnit')\nplatform_unit.moveTo(0.0, 0.0, 0.0, 50.0, 0.1)\n```\n\nSlice definitions can be loaded using the `slice_loader.load_armarx_slice`\nfunction. Default values for the proxy name will also be mapped.\n\nMore examples can be found in the `examples` folder.\n\n## Documentation\n\nSee `https://armarx.humanoids.kit.edu/python`\n",
    'author': 'Markus Grotz',
    'author_email': 'markus.grotz@kit.edu',
    'maintainer': 'Markus Grotz',
    'maintainer_email': 'markus.grotz@kit.edu',
    'url': 'http://gitlab.com/ArmarX/python3-armarx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.9,<4.0',
}


setup(**setup_kwargs)



import enum

import numpy as np
import typing as ty

from armarx_memory.aron.aron_ice_types import AronIceTypes, dtypes_dict
from armarx_memory.aron.conversion.options import ConversionOptions


def pythonic_to_aron_ice(
    value: ty.Any,
    options: ty.Optional[ConversionOptions] = None,
) -> "armarx.aron.data.dto.GenericData":
    """
    Deeply converts objects/values of pythonic types to their Aron Ice counterparts.

    :param value: A pythonic object or value.
    :param options: Conversion options.
    :return: An Aron data Ice object.
    """

    if value is None:
        return None
    if isinstance(value, str):
        return AronIceTypes.string(value)
    elif isinstance(value, bool):
        return AronIceTypes.bool(value)
    elif isinstance(value, int) or isinstance(value, np.int32):
        return AronIceTypes.int(int(value))
    elif isinstance(value, np.int64):
        return AronIceTypes.long(int(value))
    elif isinstance(value, float):
        return AronIceTypes.float(value)
    elif isinstance(value, list):
        return AronIceTypes.list(list(map(pythonic_to_aron_ice, value)))
    elif isinstance(value, enum.IntEnum):
        return pythonic_to_aron_ice(value.value)  # int

    elif isinstance(value, dict):
        a = AronIceTypes.dict(
            {
                (
                    options.name_python_to_aron(k) if options is not None else k
                ): pythonic_to_aron_ice(v)
                for k, v in value.items()
            }
        )
        return a

    elif isinstance(value, AronIceTypes.Dict):
        return value

    elif isinstance(value, np.ndarray):
        return ndarray_to_aron(value)

    try:
        return value.to_aron()
    except TypeError:
        pass

    raise TypeError(f"Could not convert object of type '{type(value)}' to aron.")


def pythonic_from_aron_ice(
    data: "armarx.aron.data.dto.GenericData",
    options: ty.Optional[ConversionOptions] = None,
) -> ty.Any:
    """
    Deeply converts an Aron data Ice object to its pythonic representation.

    :param data: The Aron data Ice object.
    :param options: Conversion options.
    :return: The pythonic representation.
    """

    def handle_dict(elements):
        return {
            (
                options.name_aron_to_python(k) if options is not None else k
            ): pythonic_from_aron_ice(v)
            for k, v in elements.items()
        }

    def handle_list(elements):
        return list(map(pythonic_from_aron_ice, elements))

    if data is None:
        return None
    if isinstance(data, list):
        return handle_list(data)
    elif isinstance(data, dict):
        return handle_dict(data)

    elif isinstance(data, (float, int, str)):
        return data

    if isinstance(data, AronIceTypes.NDArray):
        return ndarray_from_aron(data)

    try:
        return data.value
    except AttributeError:
        pass

    try:
        elements = data.elements
    except AttributeError:
        pass
    else:
        if isinstance(elements, list):
            return handle_list(elements)
        elif isinstance(elements, dict):
            return handle_dict(elements)
        else:
            raise TypeError(
                f"Could not handle aron container object of type '{type(data)}'. \n"
                f"elements: {elements}"
            )

    raise TypeError(
        f"Could not handle aron object of type '{type(data)}'.\n" f"dir(a): {dir(data)}"
    )


def ndarray_to_aron(value: np.ndarray) -> AronIceTypes.NDArray:
    shape = (*value.shape, value.itemsize)
    return AronIceTypes.NDArray(
        shape=shape, type=str(value.dtype), data=value.tobytes()
    )


def ndarray_from_aron(data: AronIceTypes.NDArray) -> np.ndarray:
    # Last entry is #bytes per entry
    byte_data: bytes = data.data

    shape: ty.Tuple[int]
    try:
        shape = data.dimensions[:-1]
    except AttributeError:
        shape = data.shape[:-1]
    shape = tuple(shape)

    dtype = dtypes_dict.get(data.type, None)
    if dtype is None:
        size = np.product(shape)
        if size == 0:
            dtype = np.uint8
        else:
            dtype_size = len(byte_data) // size
            dtype_dict = {1: np.uint8, 2: np.uint16, 4: np.uint32, 8: np.uint64}
            dtype = dtype_dict.get(dtype_size, None)
            if dtype is None:
                # Build a structured dtype with sequence of bytes.
                dtype = np.dtype([("bytes", np.uint8, dtype_size)])

        print(
            f"Unknown type '{data.type}' of array with shape {shape} and {len(byte_data)} bytes. "
            f"Falling back to {dtype}."
        )

    array: np.ndarray = np.frombuffer(buffer=byte_data, dtype=dtype)
    array = array.reshape(shape)
    return array



import logging
import threading

from armarx import TextListenerInterfacePrx

from armarx import TextToSpeechStateInterface
from armarx import TextToSpeechStateType

from armarx_core import ice_manager

logger = logging.getLogger(__name__)


class TextStateListener(TextToSpeechStateInterface):
    """
    A simple text state listener class.
    """

    def __init__(self):
        super().__init__()
        self.cv = threading.Condition()
        self.state = TextToSpeechStateType.eIdle
        self.tts = ice_manager.get_topic(TextListenerInterfacePrx, "TextToSpeech")

    def reportState(self, state, c=None):
        with self.cv:
            self.state = state
            self.cv.notify()

    def on_connect(self):
        """ """
        logger.debug("Registering TextListener")
        self._proxy = ice_manager.register_object(self, self.__class__.__name__)
        ice_manager.using_topic(self._proxy, self.__class__.__name__ + ".Listener")

    def is_idle(self) -> bool:
        """
        Returns true if the TTS system is idle
        """
        return self.state == TextToSpeechStateType.eIdle

    def say(self, text: str) -> None:
        """
        verbalize text.  The methods waits until the TTS system is idle.
        The text can be either a plain string or SSML markup.
        ..see:: https://cloud.google.com/text-to-speech/docs/ssml


        :param text: the text to report to the TTS system
        """
        self.cv.wait_for(lambda: self.is_idle(), timeout=30)
        self.tts.reportText(text)



"""
This module provides functionality for providing point clouds in ArmarX.

Classes:
- PointCloudProvider: Can provide point clouds as numpy arrays.
"""

import logging
import time

import numpy as np

from armarx_core import ice_manager

from visionx.pointclouds import dtype_point_color_xyz, get_point_cloud_format
from visionx.pointclouds import (
    PointCloudProviderInterface,
    PointCloudProcessorInterfacePrx,
)


logger = logging.getLogger(__name__)


class PointCloudProvider(PointCloudProviderInterface):
    """
    A point cloud provider offers point clouds.

    A new point cloud can be provided by update_point_cloud(). The point cloud should be created as
    a numpy array with the respective structured data type that was specified in the constructor (point_dt).
    You can use the create_point_cloud_array() method to create a compatible numpy array.
    """

    def __init__(
        self,
        name: str,
        point_dtype: np.dtype = dtype_point_color_xyz,
        initial_capacity: int = 640 * 480,
        connect: bool = False,
    ):
        super().__init__()
        self.name = name
        self.point_dtype = point_dtype
        self.format = get_point_cloud_format(initial_capacity, point_dtype)
        # The points array is pre-allocated.
        # When update_point_cloud is called, the new data is copied into this array.
        self.points = self.create_point_cloud_array(initial_capacity)

        self.pc_topic = None
        self.proxy = None

        if connect:
            self.on_connect()

    def create_point_cloud_array(self, shape):
        """
        Create a numpy array with compatible type to provide to update_point_cloud() later.

        :param shape: Shape of the array to be created.
        :return: np.array with the desired shape and compatible dtype.
        """
        return np.zeros(shape, self.point_dtype)

    def on_connect(self):
        """
        Register the point cloud provider in Ice.

        Call this function before calling update_point_cloud().
        """
        logger.debug("registering point cloud provider %s", self.name)
        self.proxy = ice_manager.register_object(self, self.name)
        self.pc_topic = ice_manager.get_topic(
            PointCloudProcessorInterfacePrx, f"{self.name}.PointCloudListener"
        )

    def on_disconnect(self):
        """
        Currently not implemented, but might be used to cleanup after the provider is no longer needed.
        """
        # Does nothing currently
        pass

    def update_point_cloud(self, points: np.ndarray, time_provided: int = 0):
        """
        Publish a new point cloud

        :param points: np.array of points with compatible dtype.
        :param time_provided: time stamp of the images. If zero the current time will be used
        """
        if points.dtype != self.point_dtype:
            raise Exception(
                "Array data type is not compatible!", points.dtype, self.point_dtype
            )

        # Do we need to guard this with a mutex? Probably...
        if points.shape[0] != self.points.shape[0]:
            self.points = np.copy(points)
        else:
            np.copyto(self.points, points)
        # format.size expects number of bytes (a point is np.float32 is 4 bytes)
        number_of_points = points.shape[0]
        new_size = number_of_points * points.dtype.itemsize

        self.format.size = new_size
        self.format.width = number_of_points
        self.format.timeProvided = time_provided or int(time.time() * 1000.0 * 1000.0)

        if self.pc_topic:
            self.pc_topic.reportPointCloudAvailable(self.name)
        else:
            logger.warning("not connected. call on_connect() method")

    def getPointCloudFormat(self, current=None):
        logger.debug("getPointCloudFormat() %s", self.format)
        return self.format

    def getPointCloud(self, current=None):
        logger.debug("getPointCloud() %s", self.format)
        return self.points, self.format

    def hasSharedMemorySupport(self, current=None):
        return False



import copy
import numpy as np

from armarx import RemoteGuiInterfacePrx
from armarx import RemoteGui as rg


def make_value_variant(py_value) -> rg.ValueVariant:
    result = rg.ValueVariant()
    if py_value is None:
        result.type = rg.ValueVariantType.VALUE_VARIANT_EMPTY
    elif isinstance(py_value, bool):
        result.type = rg.ValueVariantType.VALUE_VARIANT_BOOL
        result.i = 1 if py_value else 0
    elif isinstance(py_value, int):
        result.type = rg.ValueVariantType.VALUE_VARIANT_INT
        result.i = py_value
    elif isinstance(py_value, float):
        result.type = rg.ValueVariantType.VALUE_VARIANT_FLOAT
        result.f = py_value
    elif isinstance(py_value, str):
        result.type = rg.ValueVariantType.VALUE_VARIANT_STRING
        result.s = py_value
    elif isinstance(py_value, list):
        if len(py_value) == 3:
            if isinstance(py_value[0], float):
                # Vector3 (float)
                result.type = rg.ValueVariantType.VALUE_VARIANT_VECTOR3
                result.v = py_value
            elif isinstance(py_value[0], int):
                # Vector3 (int ==> convert to float)
                result.type = rg.ValueVariantType.VALUE_VARIANT_VECTOR3
                result.v = [float(item) for item in py_value]
            elif isinstance(py_value[0], list):
                flat_list = [item for sublist in py_value for item in sublist]
                if len(flat_list) == 16:
                    result.type = rg.ValueVariantType.VALUE_VARIANT_MATRIX4
                    result.v = flat_list
                else:
                    raise ValueError(
                        "Error while creating ValueVariant from list",
                        "Number of elements in the list:",
                        len(flat_list),
                        "Value:",
                        py_value,
                    )
            else:
                raise ValueError(
                    "Error while creating ValueVariant from list",
                    "Type of first element:",
                    type(py_value[0]),
                    "Value:",
                    py_value,
                )
    elif isinstance(py_value, np.ndarray):
        return make_value_variant(py_value.tolist())
    else:
        raise ValueError(
            "Error while creating ValueVariant from unknown type",
            "Type of value:",
            type(py_value),
            "Value:",
            py_value,
        )
    return result


def unwrap_value_variant(variant: rg.ValueVariant):
    if variant.type == rg.ValueVariantType.VALUE_VARIANT_EMPTY:
        return None
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_BOOL:
        return variant.i != 0
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_INT:
        return variant.i
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_FLOAT:
        return variant.f
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_STRING:
        return variant.s
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_VECTOR3:
        return variant.v
    elif variant.type == rg.ValueVariantType.VALUE_VARIANT_MATRIX4:
        return [variant.v[0:4], variant.v[4:8], variant.v[8:12], variant.v[12:16]]
    else:
        raise ValueError("Unknown value variant type", variant)


def are_equal(left, right):
    if type(left) != type(right):
        return False
    if isinstance(left, list):
        for l, r in zip(left, right):
            return are_equal(l, r)
    else:
        return left == right


def make_widget_state(hidden: bool = False, disabled: bool = False) -> rg.WidgetState:
    result = rg.WidgetState()
    result.hidden = hidden
    result.disabled = disabled
    return result


class TabProxy:
    def __init__(self, proxy: RemoteGuiInterfacePrx, tab_id: str):
        self.proxy = proxy

        self.tab_id = tab_id

        self.current_values = {}
        self.old_values = {}
        self.new_values = {}
        self.dirty_values = {}

        self.current_widget_states = {}
        self.old_widget_states = {}
        self.new_widget_states = {}

        self.values_changed = False
        self.widget_changed = False

    def initialize_widget(self, widget):
        widget.tab = self
        for child_widget in widget.children:
            self.initialize_widget(child_widget)

    def remove(self):
        self.proxy.removeTab(self.tab_id)

    def receive_updates(self):
        self.old_values = self.current_values
        self.current_values = self.proxy.getValues(self.tab_id)
        self.new_values = copy.deepcopy(self.current_values)
        self.values_changed = False
        self.dirty_values.clear()

        self.old_widget_states = self.current_widget_states
        self.current_widget_states = self.proxy.getWidgetStates(self.tab_id)
        self.new_widget_states = copy.deepcopy(self.current_widget_states)
        self.widget_changed = False

    def send_updates(self):
        if self.values_changed:
            self.proxy.setValues(self.tab_id, self.dirty_values)
            self.dirty_values.clear()
        if self.widget_changed:
            self.proxy.setWidgetStates(self.tab_id, self.new_widget_states)

    def set_value(self, name: str, py_value):
        # Keep track of changed values, dirty values and so on
        current_value = self.get_value(name)
        new_value_variant = make_value_variant(py_value)
        if not are_equal(current_value, py_value):
            self.values_changed = True
            self.dirty_values[name] = new_value_variant
        self.new_values[name] = new_value_variant

    def get_value(self, name: str):
        current_value_variant = self.current_values[name]
        current_value = unwrap_value_variant(current_value_variant)
        return current_value

    def has_value_changed(self, name: str):
        old_value = self.old_values.get(name, None)
        if old_value is None:
            return True
        old_value = unwrap_value_variant(old_value)
        current_value = self.get_value(name)
        return not are_equal(old_value, current_value)

    def was_button_clicked(self, name: str):
        if name not in self.old_values:
            return False

        if name not in self.new_values:
            raise Exception(
                "Button with name", name, "not found in new values", self.new_values
            )

        old_value = unwrap_value_variant(self.old_values[name])
        new_value = unwrap_value_variant(self.new_values[name])
        return new_value > old_value

    def set_hidden(self, name: str, hidden: bool):
        state = self.new_widget_states[name]
        if state.hidden != hidden:
            self.widget_changed = True
        state.hidden = hidden

    def is_hidden(self, name: str) -> bool:
        state = self.current_widget_states[name]
        return state.hidden

    def set_disabled(self, name: str, disabled: bool):
        state = self.new_widget_states[name]
        if state.disabled != disabled:
            self.widget_changed = True
        state.disabled = disabled

    def is_disabled(self, name: str) -> bool:
        state = self.current_widget_states[name]
        return state.disabled



import copy
import enum

import armarx
from armarx_core import slice_loader

slice_loader.load_armarx_slice("RobotAPI", "armem/memory.ice")
from armarx import armem

from armarx_memory.ice_conv import ice_twin
from armarx_memory.core.time import DateTimeIceConverter

date_time_conv = DateTimeIceConverter()


class MemoryID(ice_twin.IceTwin):
    def __init__(
        self,
        memory_name: str = "",
        core_segment_name: str = "",
        provider_segment_name: str = "",
        entity_name: str = "",
        timestamp_usec: int = -1,
        instance_index: int = -1,
    ):
        self.memory_name = memory_name
        self.core_segment_name = core_segment_name
        self.provider_segment_name = provider_segment_name
        self.entity_name = entity_name
        self.timestamp_usec = timestamp_usec
        self.instance_index = instance_index

    @classmethod
    def from_string(cls, string: str):
        items = string.split("/")
        # Handle escaped /'s
        i = 0
        while i < len(items):
            item = items[i]
            while len(item) > 0 and item[-1] == "\\":
                # The / causing the split was escaped. Merge the items together.
                if i < len(items):
                    items[i] = item[:-1] + "/" + items[i + 1]
                    del items[i]
            i += 1

        self = cls()
        try:
            # Set as much as possible until an index error occurs.
            self.memory_name = items[0]
            self.core_segment_name = items[1]
            self.provider_segment_name = items[2]
            self.entity_name = items[3]
            self.timestamp_usec = int(items[4])
            self.instance_index = int(items[5])
        except IndexError:
            pass

        return self

    def set_memory_id(self, id: "MemoryID"):
        self.memory_name = id.memory_name

    def set_core_segment_id(self, id: "MemoryID"):
        self.set_memory_id(id)
        self.core_segment_name = id.core_segment_name

    def set_provider_segment_id(self, id: "MemoryID"):
        self.set_core_segment_id(id)
        self.provider_segment_name = id.provider_segment_name

    def set_entity_id(self, id: "MemoryID"):
        self.set_provider_segment_id(id)
        self.entity_name = id.entity_name

    def set_snapshot_id(self, id: "MemoryID"):
        self.set_entity_id(id)
        self.timestamp_usec = id.timestamp_usec

    def set_instance_id(self, id: "MemoryID"):
        self.set_snapshot_id(id)
        self.instance_index = id.instance_index

    def with_memory_name(self, name: str) -> "MemoryID":
        c = copy.copy(self)
        c.memory_name = name
        return c

    def with_core_segment_name(self, name: str) -> "MemoryID":
        c = copy.copy(self)
        c.core_segment_name = name
        return c

    def with_provider_segment_name(self, name: str) -> "MemoryID":
        c = copy.copy(self)
        c.provider_segment_name = name
        return c

    def with_entity_name(self, name: str) -> "MemoryID":
        c = copy.copy(self)
        c.entity_name = name
        return c

    def with_timestamp(self, time_usec: int) -> "MemoryID":
        c = copy.copy(self)
        c.timestamp_usec = time_usec
        return c

    def with_instance_index(self, index: int) -> "MemoryID":
        c = copy.copy(self)
        c.instance_index = index
        return c

    def contains(self, id: "MemoryID"):
        general = self
        specific = id
        if general.memory_name == "":
            return True
        elif general.memory_name != specific.memory_name:
            return False

        if general.core_segment_name == "":
            return True
        elif general.core_segment_name != specific.core_segment_name:
            return False

        if general.provider_segment_name == "":
            return True
        elif general.provider_segment_name != specific.provider_segment_name:
            return False

        if general.entity_name == "":
            return True
        elif general.entity_name != specific.entity_name:
            return False

        if general.timestamp_usec is None or general.timestamp_usec < 0:
            return True
        elif general.timestamp_usec != specific.timestamp_usec:
            return False

        if general.instance_index < 0:
            return True
        elif general.instance_index != specific.instance_index:
            return False

    def __eq__(self, other):
        if other is None or not isinstance(other, MemoryID):
            return False

        return (
            other.memory_name == self.memory_name
            and other.core_segment_name == self.core_segment_name
            and other.provider_segment_name == self.provider_segment_name
            and other.entity_name == self.entity_name
            and other.instance_index == self.instance_index
            and other.timestamp_usec == self.timestamp_usec
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # Just use two values to make has computation more efficient.
        return hash((self.timestamp_usec, self.instance_index, self.entity_name))

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, self.__str__())

    def __str__(self):
        return "'{}'".format("/".join(map(str, self.get_set_items())))

    def get_set_items(self):
        items = [self.memory_name]

        if not self.core_segment_name:
            return items
        items.append(self.core_segment_name)

        if not self.provider_segment_name:
            return items
        items.append(self.provider_segment_name)

        if not self.entity_name:
            return items
        items.append(self.entity_name)

        if self.timestamp_usec < 0:
            return items
        items.append(self.timestamp_usec)

        if self.instance_index < 0:
            return items
        items.append(self.instance_index)

        return items

    def get_all_items(self):
        return [
            self.memory_name,
            self.core_segment_name,
            self.provider_segment_name,
            self.entity_name,
            self.timestamp_usec,
            self.instance_index,
        ]

    def _get_ice_cls(self):
        return armem.data.MemoryID

    def _set_to_ice(self, ice: "armem.data.Commit"):
        ice.memoryName = self.memory_name
        ice.coreSegmentName = self.core_segment_name
        ice.providerSegmentName = self.provider_segment_name
        ice.entityName = self.entity_name
        ice.timestamp = date_time_conv.to_ice(self.timestamp_usec)
        ice.instanceIndex = self.instance_index

    def _set_from_ice(self, ice):
        self.memory_name = ice.memoryName
        self.core_segment_name = ice.coreSegmentName
        self.provider_segment_name = ice.providerSegmentName
        self.entity_name = ice.entityName
        self.timestamp_usec = date_time_conv.from_ice(ice.timestamp)
        self.instance_index = ice.instanceIndex

    @classmethod
    def from_aron(cls, aron: "armarx.aron.data.dto.GenericData") -> "MemoryID":
        from armarx_memory.aron.conversion import from_aron

        data = from_aron(aron)
        self = cls()
        self.memory_name = data["memoryName"]
        self.core_segment_name = data["coreSegmentName"]
        self.provider_segment_name = data["providerSegmentName"]
        self.entity_name = data["entityName"]
        self.timestamp_usec = int(data["timestamp"])
        self.instance_index = data["instanceIndex"]
        return self

    def to_aron(self) -> "armarx.aron.data.dto.GenericData":
        import numpy as np
        from armarx_memory.aron.conversion import to_aron

        class ClockType(enum.IntEnum):
            REALTIME = 0  # Normalized time as reported by the operating system.
            MONOTONIC = 1  # Monotonic/steady clock of the operating system.
            VIRTUAL = 2  # Time given by time-server if configured, realtime otherwise.
            UNKNOWN = 3  # Unknown source of time.

        data = {
            "memoryName": self.memory_name,
            "coreSegmentName": self.core_segment_name,
            "providerSegmentName": self.provider_segment_name,
            "entityName": self.entity_name,
            "timestamp": {
                "clockType": ClockType.VIRTUAL,
                "hostname": "unknown",
                "timeSinceEpoch": {
                    "microSeconds": np.int64(self.timestamp_usec),
                },
            },
            "instanceIndex": self.instance_index,
        }
        return to_aron(data)



import abc
from typing import Callable, Dict, Type


class IceConverter(abc.ABC):
    """
    Allows to define the conversion between a Python Business Object (BO; i.e. of a python type)
    and an Ice Data Transfer Object (DTO; i.e. an object of classes defined in Ice)
    and provides additional handling for containers of that type.

    Usage:

    class MyClassConv(IceConverter):  # Derive from IceConverter

        # Override _protected functions

        @classmethod
        def _import_dto(cls):
            slice_loader.load_armarx_slice(...)
            return armarx.dto.MyClass

        def _from_ice(self, dto: armarx.dto.MyClass) -> MyClass:
            return MyClass(data=dto.data)

        def _to_ice(self, bo: MyClass) -> armarx.dto.MyClass:
            return MyClass(data=bo.data)


    def foo():
        # Get an Ice DTO from somewhere.
        dto: armarx.dto.MyClass = get_ice_dto()

        # Create the converter and use it to convert the DTO to a BO.
        conv = MyClassConv()
        bo: MyClass = conv.from_ice(dto)

        # Use the same converter to convert the BO back to Ice.
        dto2: armarx.dto.MyClass = conv.to_ice(bo)

        # Use the same methods to convert lists or dictionaries:
        bos = conv.from_ice([dto, dto2])
        ...
    """

    _dto = None

    def __init__(self):
        self._handlers_from_ice: Dict[Type, Callable] = dict()
        self._handlers_to_ice: Dict[Type, Callable] = dict()

        self.set_handler_from_ice(
            list,
            lambda dto, *args, **kwargs: list(
                map(lambda e: self.from_ice(e, *args, **kwargs), dto)
            ),
        )
        self.set_handler_to_ice(
            list,
            lambda bo, *args, **kwargs: list(
                map(lambda e: self.to_ice(e, *args, **kwargs), bo)
            ),
        )

        self.set_handler_from_ice(
            dict,
            lambda dto, *args, **kwargs: {
                k: self.from_ice(v, *args, **kwargs) for k, v in dto.items()
            },
        )
        self.set_handler_to_ice(
            dict,
            lambda bo, *args, **kwargs: {
                k: self.to_ice(v, *args, **kwargs) for k, v in bo.items()
            },
        )

    def from_ice(self, dto, *args, **kwargs):
        """Convert the Ice Data Transfer Object(s) to Python Business Object(s)."""
        for t, handler in self._handlers_from_ice.items():
            if isinstance(dto, t):
                return handler(dto, *args, **kwargs)
        return self._from_ice(dto, *args, **kwargs)

    def to_ice(self, bo, *args, **kwargs):
        """Convert the Python Business Object(s) to Ice Data Transfer Object(s)."""
        for t, handler in self._handlers_from_ice.items():
            if isinstance(bo, t):
                return handler(bo, *args, **kwargs)
        return self._to_ice(bo, *args, **kwargs)

    def set_handler_from_ice(self, type: Type, handler: Callable):
        self._handlers_from_ice[type] = handler

    def set_handler_to_ice(self, type: Type, handler: Callable):
        self._handlers_to_ice[type] = handler

    @classmethod
    def get_dto(cls):
        if cls._dto is None:
            # Dunno why yet, but we need to pass cls explicitly here ...
            try:
                cls._dto = cls._import_dto(cls)
            except TypeError:
                cls._dto = cls._import_dto()
        return cls._dto

    @abc.abstractmethod
    def _from_ice(self, dto, *args, **kwargs):
        """Convert the Ice DTO to a Python BO."""
        ...

    @abc.abstractmethod
    def _to_ice(self, bo, *args, **kwargs):
        """Convert the Python BO to an Ice DTO."""
        ...

    @classmethod
    @abc.abstractmethod
    def _import_dto(cls):
        """Load, import and return the DTO type."""
        ...



"""
.. deprecated:: 0.20.1
    module is moved to .. py:module::`armarx_core.slice_loader` package
"""
import sys
import warnings

from armarx_core.slice_loader import ArmarXProxyFinder


def load_armarx_slice(armarx_package_name: str, filename: str):
    """
    .. deprecated:: 0.10.6

        add your slice file to a project's VariantInfo-*.xml instead
    """
    warnings.warn(
        "Add the slice definition to VariantInfo-*.xml instead.", DeprecationWarning
    )
    for c in sys.meta_path:
        if isinstance(c, ArmarXProxyFinder):
            c._load_armarx_slice(armarx_package_name, filename)
            c.update_loaded_modules()



import typing as ty

import logging

from .options import ConversionOptions


def dataclass_to_aron_ice(
    obj,
    options: ty.Optional[ConversionOptions] = None,
    logger: ty.Optional[logging.Logger] = None,
) -> "armarx.aron.data.dto.GenericData":
    from .dataclass_from_to_pythonic import dataclass_to_dict
    from .pythonic_from_to_aron_ice import pythonic_to_aron_ice

    data = dataclass_to_dict(obj, logger=logger)
    aron = pythonic_to_aron_ice(data, options=options)
    return aron


def dataclass_from_aron_ice(
    cls,
    aron: "armarx.aron.data.dto.GenericData",
    options: ty.Optional[ConversionOptions] = None,
    logger: ty.Optional[logging.Logger] = None,
):
    from .dataclass_from_to_pythonic import dataclass_from_dict
    from .pythonic_from_to_aron_ice import pythonic_from_aron_ice

    data = pythonic_from_aron_ice(aron, options=options)
    obj = dataclass_from_dict(cls, data, logger=logger)
    return obj



from typing import List, Union, Tuple, Any

import numpy as np

from armarx.viz.data import Element


import armarx.viz as viz

from armarx.arviz import conversions as conv
from .Element import Element


class Mesh(Element):
    """
    A 3D mesh made of triangular faces.

    A mesh is specified by:

    * an array `vertices` of 3D vertex positions (x, y, z) with shape (V, 3),
    * an array `colors` of colors (r, g, b) or (r, g, b, a) with shape (C, 3) or (C, 4), and
    * an array `faces` of shape (F, 6), where each entry (v1, v2, v3, c1, c2, c3) specifies
      the indices of the three adjacent vertices (indexing the `vertices` array) and indices
      of adjacent vertex colors (indexing the `colors` array).

    """

    def __init__(
        self,
        id,
        **kwargs,
    ):
        super().__init__(ice_data_cls=viz.data.ElementMesh, id=id, **kwargs)

        self.vertices = []
        self.colors = []
        self.faces = []

    @property
    def vertices(self) -> np.ndarray:
        """An N x 3 array of vertex positions."""
        assert self._match_shapes(
            self._vertices.shape, [(0,), (None, 3)]
        ), self._vertices.shape
        return self._vertices

    @vertices.setter
    def vertices(self, value):
        value = self._to_array_checked(
            value, [(0,), (None, 3)], "mesh vertices", dtype=np.float
        )
        self._vertices = value

    @property
    def colors(self):
        """An N x 4 array of vertex colors."""
        assert self._match_shapes(
            self._colors.shape, [(0,), (None, 4)]
        ), self._colors.shape
        return self._colors

    @colors.setter
    def colors(self, value):
        value = self._to_array_checked(
            value, [(0,), (None, 3), (None, 4)], "mesh colors", dtype=np.float
        )
        if value.shape[-1] == 3:
            self._colors = np.concatenate([value, [[255]] * len(value)], axis=-1)
        else:
            self._colors = value

    @property
    def faces(self) -> np.ndarray:
        """
        An N x 6 array of vertex position and color indices.
        An entry has the form:
        [ v0 v1 v2 c0 c1 c2 ]
        """
        assert self._match_shapes(self._faces.shape, [(0,), (None, 6)])
        return self._faces

    @faces.setter
    def faces(self, value: Union[np.ndarray, List[viz.data.Face]]):
        value = self._to_array_checked(
            value, [(0,), (None, 6)], "mesh faces", dtype=int
        )
        self._faces = value

    def _update_ice_data(self, ice_data):
        super()._update_ice_data(ice_data)

        ice_data.vertices = conv.vector3fs_from_numpy(self.vertices)
        ice_data.colors = [conv.to_viz_color(c) for c in self.colors]
        ice_data.faces = [viz.data.Face(*map(int, f)) for f in self.faces]

    @staticmethod
    def make_grid2d_faces(
        num_x: int,
        num_y: int,
    ) -> np.ndarray:
        """
        Make vertex and color indices of triangular faces of a 2D grid.

        Note that the vertex and color indices are independent of concrete
        vertex positions and color values. They merely state which vertices
        are connected to triangular faces and which color each vertex (in each
        face) has. The vertex positions and color values are specified by
        the `vertices` and `colors` arrays.

        Indices created by this function match the order of vertices
        produced by `armarx.arviz.grids.make_grid()` and related functions.

        :param num_x:
            The number of vertices in first dimension.
        :param num_y:
            The number of vertices in second dimension.
        :return:
            An array f of shape (num_faces, 3+3) with f[i] = (v1, v2, v3, c1, c2, c3)
            are the vertex and color indices of a triangular face.
        """
        num_faces = 2 * (num_x - 1) * (num_y - 1)
        faces = np.zeros((num_faces, 6))

        i = 0
        for x in range(num_x - 1):
            for y in range(num_y - 1):
                """
                In counter-clockwise order:
                      (x)  (x+1)
                 (y)   *----*
                       | \f1|
                       |f2\ |
                 (y+1) *----*
                """
                v0 = c0 = y * num_x + x
                v1 = c1 = (y + 1) * num_x + (x + 1)
                v2 = c2 = (y + 1) * num_x + x
                faces[i] = v0, v1, v2, c0, c1, c2
                i += 1

                v0 = c0 = y * num_x + x
                v1 = c1 = y * num_x + (x + 1)
                v2 = c2 = (y + 1) * num_x + (x + 1)
                faces[i] = v0, v1, v2, c0, c1, c2
                i += 1

        assert i == num_faces

        return faces



import logging

from typing import List, Union

from armarx import ice_manager

from armarx.arviz.layer import Layer
from armarx.arviz.stage import Stage
from armarx.arviz.interaction_feedback import CommitResult
from armarx.viz import StorageInterfacePrx
import armarx.viz as viz

logger = logging.getLogger(__name__)


class Client:
    """
    An ArViz client.
    """

    STORAGE_DEFAULT_NAME = "ArVizStorage"

    def __init__(
        self,
        component: str,
        storage_name=STORAGE_DEFAULT_NAME,
        wait_for_proxy=True,
    ):
        self.component_name = component

        args = (viz.StorageInterfacePrx, storage_name)
        if wait_for_proxy:
            self.storage = ice_manager.wait_for_proxy(*args)
        else:
            self.storage = ice_manager.get_proxy(*args)

    def layer(self, name) -> Layer:
        """
        Create a layer.
        :param name: The layer's name.
        :return: The layer.
        """
        return Layer(self.component_name, name)

    def begin_stage(self, commit_on_exit=False) -> Stage:
        if commit_on_exit:
            return Stage(
                self.component_name, commit_on_exit=commit_on_exit, client=self
            )
        else:
            return Stage(self.component_name)

    def commit(
        self,
        layers_or_stages: Union[None, Layer, Stage, List[Union[Layer, Stage]]] = None,
    ) -> CommitResult:
        """
        Commit the given layers and stages.
        :param layers_or_stages: Layer(s) or Stage(s) to commit.
        """
        if layers_or_stages is None:
            layers_or_stages = []
        try:
            iter(layers_or_stages)
        except TypeError:
            # Single item.
            layers_or_stages = [layers_or_stages]

        input_ = viz.data.CommitInput()
        input_.updates = sum(map(self._get_layer_updates, layers_or_stages), [])
        input_.interactionComponent = self.component_name

        interaction_layers: List[str] = []

        for layer_or_stage in layers_or_stages:
            if isinstance(layer_or_stage, Stage):
                stage = layer_or_stage
                interaction_layers += stage._interaction_layers

        input_.interactionLayers = interaction_layers

        ice_result = self.storage.commitAndReceiveInteractions(input_)

        result = CommitResult(data=ice_result)
        return result

    @staticmethod
    def _get_layer_updates(
        layer_like: Union[Layer, viz.data.LayerUpdate],
    ) -> List[viz.data.LayerUpdate]:

        if isinstance(layer_like, viz.data.LayerUpdate):
            return [layer_like]
        elif isinstance(layer_like, Layer):
            return [layer_like.data()]
        elif isinstance(layer_like, Stage):
            return [layer.data() for layer in layer_like.layers]
        else:
            logger.warning("Unable to get layer updates.")
            return []



import typing as ty
import logging

from armarx_core import slice_loader
from armarx_core import ice_manager

slice_loader.load_armarx_slice("RobotAPI", "armem/client/MemoryListenerInterface.ice")

from armarx import armem

from armarx_memory.core.MemoryID import MemoryID


logger = logging.getLogger(__file__)


class MemoryListener(armem.client.MemoryListenerInterface):

    Callback = ty.Callable[[MemoryID, ty.List[MemoryID]], None]
    UpdatedSnapshotIDs = ty.List[ty.Union[MemoryID, "armarx.armem.data.MemoryID"]]

    TopicNameFormat = "MemoryUpdates.{memory_name}"

    def __init__(
        self,
        name: ty.Optional[str] = None,
        register=True,
        log_fn=None,
    ):
        self.name = name
        self.proxy = None

        if log_fn is None:

            def log_fn(*args, **kwargs):
                pass

        self.log_fn = log_fn

        self.subscriptions: ty.Dict[MemoryID, ty.List["MemoryListener.Callback"]] = {}

        if register:
            self.register()

    def register(self):
        self.log_fn(f"Register {self.__class__.__name__} '{self.name}' ...")

        self.proxy = ice_manager.register_object(self, self.name)
        return self.proxy

    def use_topic_of_id(self, memory_id: MemoryID):
        topic_name = self.TopicNameFormat.format(memory_name=memory_id.memory_name)
        self.log_fn(f"'{self.name}': Use topic '{topic_name}'.")
        ice_manager.using_topic(self.proxy, topic_name)

    def subscribe(self, subscription_id: MemoryID, callback: Callback):
        """
        Subscribe a memory ID in order to receive updates to it.
        :param subscription_id: The subscribed ID.
        :param callback: The callback to be called with the updated IDs.
        """
        self.log_fn(f"'{self.name}': Subscribe to {subscription_id}.")
        self.use_topic_of_id(memory_id=subscription_id)
        if subscription_id not in self.subscriptions:
            self.subscriptions[subscription_id] = [callback]
        else:
            self.subscriptions[subscription_id].append(callback)

    def updated(self, updated_snapshot_ids: UpdatedSnapshotIDs):
        """
        Function to be called when receiving messages over MemoryListener topic.
        :param updated_snapshot_ids: The updated snapshot IDs.
        """
        # Convert from Ice
        updated_snapshot_ids: ty.List[MemoryID] = [
            id if isinstance(id, MemoryID) else MemoryID.from_ice(id)
            for id in updated_snapshot_ids
        ]
        for id in updated_snapshot_ids:
            assert isinstance(id, MemoryID)

        for subscribed_id, callbacks in self.subscriptions.items():
            # Split by subscribed id
            matching_snapshot_ids = [
                updated_snapshot_id
                for updated_snapshot_id in updated_snapshot_ids
                if subscribed_id.contains(updated_snapshot_id)
            ]
            # Call callbacks
            if len(matching_snapshot_ids) > 0:
                for callback in callbacks:
                    callback(subscribed_id, matching_snapshot_ids)

    def memoryUpdated(self, updated_snapshot_ids: ty.List[armem.data.MemoryID], c=None):
        """Called via the MemoryListenerTopic."""
        updated_snapshot_ids = MemoryID.from_ice(updated_snapshot_ids)
        self.updated(updated_snapshot_ids)
